import copy
import logging
import mock
import pytest
import requests_mock
import requests

from pubtools._quay import exceptions
from pubtools._quay import quay_client
from pubtools._quay import container_image_pusher
from .utils.misc import sort_dictionary_sortable_values, compare_logs

# flake8: noqa: E501


@mock.patch("pubtools._quay.container_image_pusher.QuayClient")
def test_init(mock_quay_client, target_settings, container_multiarch_push_item):
    pusher = container_image_pusher.ContainerImagePusher(
        [container_multiarch_push_item], target_settings
    )

    assert pusher.push_items == [container_multiarch_push_item]
    assert pusher.target_settings == target_settings
    assert pusher.quay_host == "quay.io"
    mock_quay_client.assert_not_called()

    assert pusher.src_quay_client == mock_quay_client.return_value
    assert pusher.dest_quay_client == mock_quay_client.return_value
    assert mock_quay_client.call_args_list == [
        mock.call("src-quay-user", "src-quay-pass", "quay.io"),
        mock.call("dest-quay-user", "dest-quay-pass", "quay.io"),
    ]


@mock.patch("pubtools._quay.container_image_pusher.tag_images")
@mock.patch("pubtools._quay.container_image_pusher.QuayClient")
def test_tag_images(
    mock_quay_client,
    mock_tag_images,
    target_settings,
    container_multiarch_push_item,
):
    pusher = container_image_pusher.ContainerImagePusher(
        [container_multiarch_push_item], target_settings
    )

    pusher.run_tag_images(
        "source-ref:1", ["dest-ref:1", "dest-ref:2"], True, pusher.target_settings
    )
    mock_tag_images.assert_called_once_with(
        "source-ref:1",
        ["dest-ref:1", "dest-ref:2"],
        all_arch=True,
        quay_user="dest-quay-user",
        quay_password="dest-quay-pass",
        source_quay_host=None,
        source_quay_user="src-quay-user",
        source_quay_password="src-quay-pass",
        container_exec=True,
        container_image="registry.com/some/image:1",
        docker_url="unix://var/run/docker.sock",
        docker_timeout=None,
        docker_verify_tls=False,
        docker_cert_path=None,
        registry_username="quay-executor-user",
        registry_password="quay-executor-password",
    )


@mock.patch("time.sleep")
@mock.patch("pubtools._quay.container_image_pusher.tag_images")
@mock.patch("pubtools._quay.container_image_pusher.QuayClient")
def test_tag_images_retry(
    mock_quay_client,
    mock_tag_images,
    mock_sleep,
    target_settings,
    container_multiarch_push_item,
):
    mock_tag_images.side_effect = [RuntimeError, ValueError, 0]

    pusher = container_image_pusher.ContainerImagePusher(
        [container_multiarch_push_item], target_settings
    )

    pusher.run_tag_images(
        "source-ref:1", ["dest-ref:1", "dest-ref:2"], True, pusher.target_settings
    )

    tag_images_call = mock.call(
        "source-ref:1",
        ["dest-ref:1", "dest-ref:2"],
        all_arch=True,
        quay_user="dest-quay-user",
        quay_password="dest-quay-pass",
        source_quay_host=None,
        source_quay_user="src-quay-user",
        source_quay_password="src-quay-pass",
        container_exec=True,
        container_image="registry.com/some/image:1",
        docker_url="unix://var/run/docker.sock",
        docker_timeout=None,
        docker_verify_tls=False,
        docker_cert_path=None,
        registry_username="quay-executor-user",
        registry_password="quay-executor-password",
    )

    assert mock_tag_images.call_count == 3
    assert mock_tag_images.call_args_list[0] == tag_images_call
    assert mock_tag_images.call_args_list[1] == tag_images_call
    assert mock_tag_images.call_args_list[2] == tag_images_call


@mock.patch("pubtools._quay.container_image_pusher.tag_images")
@mock.patch("pubtools._quay.container_image_pusher.QuayClient")
def test_copy_src_item(
    mock_quay_client,
    mock_tag_images,
    target_settings,
    container_source_push_item,
):
    pusher = container_image_pusher.ContainerImagePusher(
        [container_source_push_item], target_settings
    )
    pusher.copy_source_push_item(container_source_push_item)
    mock_tag_images.assert_called_once_with(
        "some-registry/src/repo:2",
        [
            "quay.io/some-namespace/target----repo:latest-test-tag",
            "quay.io/some-namespace/target----repo:1.0",
        ],
        all_arch=True,
        quay_user="dest-quay-user",
        quay_password="dest-quay-pass",
        container_exec=True,
        container_image="registry.com/some/image:1",
        docker_url="unix://var/run/docker.sock",
        docker_timeout=None,
        docker_verify_tls=False,
        docker_cert_path=None,
        registry_username="quay-executor-user",
        registry_password="quay-executor-password",
        source_quay_host=None,
        source_quay_password="src-quay-pass",
        source_quay_user="src-quay-user",
    )


@mock.patch("pubtools._quay.container_image_pusher.tag_images")
@mock.patch("pubtools._quay.container_image_pusher.QuayClient")
def test_copy_v1_item(
    mock_quay_client,
    mock_tag_images,
    target_settings,
    container_v1_push_item,
):
    pusher = container_image_pusher.ContainerImagePusher([container_v1_push_item], target_settings)
    pusher.copy_v1_push_item(container_v1_push_item)
    mock_tag_images.assert_called_once_with(
        "some-registry/src/repo:2",
        [
            "quay.io/some-namespace/target----repo:latest-test-tag",
            "quay.io/some-namespace/target----repo:1.0",
        ],
        all_arch=True,
        quay_user="dest-quay-user",
        quay_password="dest-quay-pass",
        source_quay_host=None,
        source_quay_user="src-quay-user",
        source_quay_password="src-quay-pass",
        container_exec=True,
        container_image="registry.com/some/image:1",
        docker_url="unix://var/run/docker.sock",
        docker_timeout=None,
        docker_verify_tls=False,
        docker_cert_path=None,
        registry_username="quay-executor-user",
        registry_password="quay-executor-password",
    )


@mock.patch("pubtools._quay.container_image_pusher.ManifestListMerger")
@mock.patch("pubtools._quay.container_image_pusher.tag_images")
@mock.patch("pubtools._quay.container_image_pusher.QuayClient")
def test_merge_workflow(
    mock_quay_client,
    mock_tag_images,
    mock_ml_merger,
    target_settings,
    container_multiarch_push_item,
):
    mock_get_manifest = mock.MagicMock()
    mock_get_manifest.return_value = {"manifests": [{"digest": "digest1"}, {"digest": "digest2"}]}
    mock_quay_client.return_value.get_manifest = mock_get_manifest

    pusher = container_image_pusher.ContainerImagePusher(
        [container_multiarch_push_item], target_settings
    )
    pusher.run_merge_workflow(
        "registry/src/image:1", ["registry/dest1/image:1", "registry/dest2/image:2"]
    )
    mock_get_manifest.assert_called_once_with(
        "registry/src/image:1", media_type=mock_quay_client.MANIFEST_LIST_TYPE
    )
    # test that src digests are copied to all dest repos
    assert mock_tag_images.call_args_list[0][0][1] == [
        "registry/dest1/image@digest1",
        "registry/dest2/image@digest1",
    ]
    assert mock_tag_images.call_args_list[1][0][1] == [
        "registry/dest1/image@digest2",
        "registry/dest2/image@digest2",
    ]

    assert mock_ml_merger.call_args_list == [
        mock.call("registry/src/image:1", "registry/dest1/image:1", host="quay.io"),
        mock.call("registry/src/image:1", "registry/dest2/image:2", host="quay.io"),
    ]

    assert len(mock_ml_merger.mock_calls) == 6


@mock.patch("pubtools._quay.container_image_pusher.ContainerImagePusher.run_merge_workflow")
@mock.patch("pubtools._quay.container_image_pusher.ManifestListMerger.get_missing_architectures")
@mock.patch("pubtools._quay.container_image_pusher.tag_images")
@mock.patch("pubtools._quay.container_image_pusher.QuayClient")
def test_copy_multiarch_item_no_extra_archs(
    mock_quay_client,
    mock_tag_images,
    mock_get_missing_archs,
    mock_merge_workflow,
    target_settings,
    container_multiarch_push_item,
):
    mock_get_manifest = mock.MagicMock()
    mock_get_manifest.return_value = {"manifest_list": "second_ml"}
    mock_quay_client.return_value.get_manifest = mock_get_manifest
    mock_get_missing_archs.return_value = []

    pusher = container_image_pusher.ContainerImagePusher(
        [container_multiarch_push_item], target_settings
    )
    pusher.copy_multiarch_push_item(container_multiarch_push_item, {"manifest_list": "first_ml"})

    mock_get_manifest.assert_called_once_with(
        "quay.io/some-namespace/target----repo:latest-test-tag"
    )
    assert mock_tag_images.call_count == 1
    assert mock_tag_images.call_args_list[0][0] == (
        "some-registry/src/repo:1",
        ["quay.io/some-namespace/target----repo:latest-test-tag"],
    )


@mock.patch("pubtools._quay.container_image_pusher.ContainerImagePusher.run_merge_workflow")
@mock.patch("pubtools._quay.container_image_pusher.ManifestListMerger.get_missing_architectures")
@mock.patch("pubtools._quay.container_image_pusher.tag_images")
@mock.patch("pubtools._quay.container_image_pusher.QuayClient")
def test_copy_multiarch_item_no_dest_ml(
    mock_quay_client,
    mock_tag_images,
    mock_get_missing_archs,
    mock_merge_workflow,
    target_settings,
    container_multiarch_push_item,
):
    mock_get_manifest = mock.MagicMock()

    response = mock.MagicMock()
    response.status_code = 401
    mock_get_manifest.side_effect = requests.exceptions.HTTPError("some error", response=response)
    mock_quay_client.return_value.get_manifest = mock_get_manifest
    mock_get_missing_archs.return_value = []

    pusher = container_image_pusher.ContainerImagePusher(
        [container_multiarch_push_item], target_settings
    )
    pusher.copy_multiarch_push_item(
        container_multiarch_push_item,
        {"manifest_list": "first_ml"},
    )

    mock_get_manifest.assert_called_once_with(
        "quay.io/some-namespace/target----repo:latest-test-tag",
    )

    assert mock_tag_images.call_count == 1
    assert mock_tag_images.call_args_list[0][0] == (
        "some-registry/src/repo:1",
        ["quay.io/some-namespace/target----repo:latest-test-tag"],
    )

    mock_merge_workflow.assert_not_called()


@mock.patch("pubtools._quay.container_image_pusher.ContainerImagePusher.run_merge_workflow")
@mock.patch("pubtools._quay.container_image_pusher.ManifestListMerger.get_missing_architectures")
@mock.patch("pubtools._quay.container_image_pusher.tag_images")
@mock.patch("pubtools._quay.container_image_pusher.QuayClient")
def test_copy_multiarch_item_network_error(
    mock_quay_client,
    mock_tag_images,
    mock_get_missing_archs,
    mock_merge_workflow,
    target_settings,
    container_multiarch_push_item,
):
    mock_get_manifest = mock.MagicMock()

    response = mock.MagicMock()
    response.status_code = 500
    mock_get_manifest.side_effect = requests.exceptions.HTTPError("bad error", response=response)

    mock_quay_client.return_value.get_manifest = mock_get_manifest
    mock_get_missing_archs.return_value = []

    pusher = container_image_pusher.ContainerImagePusher(
        [container_multiarch_push_item], target_settings
    )
    with pytest.raises(requests.exceptions.HTTPError, match="bad error"):
        pusher.copy_multiarch_push_item(
            container_multiarch_push_item, {"manifest_list": "first_ml"}
        )


@mock.patch("pubtools._quay.container_image_pusher.ContainerImagePusher.run_merge_workflow")
@mock.patch("pubtools._quay.container_image_pusher.ManifestListMerger.get_missing_architectures")
@mock.patch("pubtools._quay.container_image_pusher.tag_images")
@mock.patch("pubtools._quay.container_image_pusher.QuayClient")
def test_copy_multiarch_item_missing_archs(
    mock_quay_client,
    mock_tag_images,
    mock_get_missing_archs,
    mock_merge_workflow,
    target_settings,
    container_multiarch_push_item,
):
    mock_get_manifest = mock.MagicMock()

    mock_get_manifest.return_value = {
        "manifest_list": "second_ml",
        "mediaType": "application/vnd.docker.distribution.manifest.list.v2+json",
    }
    mock_quay_client.MANIFEST_LIST_TYPE = (
        "application/vnd.docker.distribution.manifest.list.v2+json"
    )
    mock_quay_client.return_value.get_manifest = mock_get_manifest
    mock_get_missing_archs.return_value = [{"arch": "x86_64"}, {"arch": "amd64"}]

    pusher = container_image_pusher.ContainerImagePusher(
        [container_multiarch_push_item], target_settings
    )
    pusher.copy_multiarch_push_item(container_multiarch_push_item, {"manifest_list": "first_ml"})

    mock_get_manifest.assert_called_once_with(
        "quay.io/some-namespace/target----repo:latest-test-tag",
    )

    assert mock_merge_workflow.call_count == 1
    assert mock_merge_workflow.call_args_list[0][0] == (
        "some-registry/src/repo:1",
        ["quay.io/some-namespace/target----repo:latest-test-tag"],
    )

    mock_tag_images.assert_not_called()


@mock.patch("pubtools._quay.container_image_pusher.ContainerImagePusher.run_merge_workflow")
@mock.patch("pubtools._quay.container_image_pusher.ManifestListMerger.get_missing_architectures")
@mock.patch("pubtools._quay.container_image_pusher.tag_images")
@mock.patch("pubtools._quay.container_image_pusher.QuayClient")
def test_copy_multiarch_item_existing_dest_not_manifest_list(
    mock_quay_client,
    mock_tag_images,
    mock_get_missing_archs,
    mock_merge_workflow,
    target_settings,
    container_multiarch_push_item,
    caplog,
):
    caplog.set_level(logging.WARNING)
    mock_get_manifest = mock.MagicMock()

    mock_get_manifest.return_value = {
        "manifest_list": "v2s2_manifest",
        "mediaType": "application/vnd.docker.distribution.manifest.v2+json",
    }
    mock_quay_client.MANIFEST_LIST_TYPE = (
        "application/vnd.docker.distribution.manifest.list.v2+json"
    )
    mock_quay_client.return_value.get_manifest = mock_get_manifest

    pusher = container_image_pusher.ContainerImagePusher(
        [container_multiarch_push_item], target_settings
    )
    pusher.copy_multiarch_push_item(container_multiarch_push_item, {"manifest_list": "first_ml"})

    mock_get_manifest.assert_called_once_with(
        "quay.io/some-namespace/target----repo:latest-test-tag",
    )

    mock_merge_workflow.assert_not_called()
    assert mock_tag_images.call_count == 1
    assert mock_tag_images.call_args_list[0][0] == (
        "some-registry/src/repo:1",
        ["quay.io/some-namespace/target----repo:latest-test-tag"],
    )

    expected_logs = [
        "Image quay.io/some-namespace/target----repo:latest-test-tag doesn't have a manifest.*",
    ]
    compare_logs(caplog, expected_logs)


@mock.patch("pubtools._quay.container_image_pusher.ContainerImagePusher.copy_source_push_item")
@mock.patch("pubtools._quay.container_image_pusher.ContainerImagePusher.copy_multiarch_push_item")
@mock.patch("pubtools._quay.container_image_pusher.QuayClient")
def test_push_container_items_src_item(
    mock_quay_client,
    mock_copy_multiarch,
    mock_copy_src,
    target_settings,
    container_source_push_item,
):
    mock_get_manifest = mock.MagicMock()

    mock_get_manifest.side_effect = exceptions.ManifestTypeError("no manifest list")
    mock_quay_client.return_value.get_manifest = mock_get_manifest

    pusher = container_image_pusher.ContainerImagePusher(
        [container_source_push_item], target_settings
    )
    pusher.push_container_images()

    mock_copy_multiarch.assert_not_called()
    mock_copy_src.assert_called_once()


@mock.patch("pubtools._quay.container_image_pusher.ContainerImagePusher.copy_source_push_item")
@mock.patch("pubtools._quay.container_image_pusher.ContainerImagePusher.copy_multiarch_push_item")
@mock.patch("pubtools._quay.container_image_pusher.QuayClient")
def test_push_container_items_src_item_external_registry(
    mock_quay_client,
    mock_copy_multiarch,
    mock_copy_src,
    target_settings,
    container_source_push_item,
):
    mock_get_manifest = mock.MagicMock()

    response_404 = requests.Response()
    response_404.status_code = 404
    mock_get_manifest.side_effect = requests.exceptions.HTTPError(response=response_404)
    mock_quay_client.return_value.get_manifest = mock_get_manifest

    pusher = container_image_pusher.ContainerImagePusher(
        [container_source_push_item], target_settings
    )
    pusher.push_container_images()

    mock_copy_multiarch.assert_not_called()
    mock_copy_src.assert_called_once()


@mock.patch("pubtools._quay.container_image_pusher.ContainerImagePusher.copy_source_push_item")
@mock.patch("pubtools._quay.container_image_pusher.ContainerImagePusher.copy_multiarch_push_item")
@mock.patch("pubtools._quay.container_image_pusher.QuayClient")
def test_push_container_items_src_item_500_error(
    mock_quay_client,
    mock_copy_multiarch,
    mock_copy_src,
    target_settings,
    container_source_push_item,
):
    mock_get_manifest = mock.MagicMock()

    response_500 = requests.Response()
    response_500.status_code = 500
    mock_get_manifest.side_effect = requests.exceptions.HTTPError(response=response_500)
    mock_quay_client.return_value.get_manifest = mock_get_manifest

    pusher = container_image_pusher.ContainerImagePusher(
        [container_source_push_item], target_settings
    )
    with pytest.raises(requests.exceptions.HTTPError):
        pusher.push_container_images()


@mock.patch("pubtools._quay.container_image_pusher.ContainerImagePusher.copy_v1_push_item")
@mock.patch("pubtools._quay.container_image_pusher.ContainerImagePusher.copy_multiarch_push_item")
@mock.patch("pubtools._quay.container_image_pusher.QuayClient")
def test_push_container_items_v1_item(
    mock_quay_client,
    mock_copy_multiarch,
    mock_copy_src,
    target_settings,
    container_v1_push_item,
):
    mock_get_manifest = mock.MagicMock()

    mock_get_manifest.side_effect = exceptions.ManifestTypeError("no manifest list")
    mock_quay_client.return_value.get_manifest = mock_get_manifest

    pusher = container_image_pusher.ContainerImagePusher([container_v1_push_item], target_settings)
    pusher.push_container_images()

    mock_copy_multiarch.assert_not_called()
    mock_copy_src.assert_called_once()


@mock.patch("pubtools._quay.container_image_pusher.ContainerImagePusher.copy_v1_push_item")
@mock.patch("pubtools._quay.container_image_pusher.ContainerImagePusher.copy_multiarch_push_item")
@mock.patch("pubtools._quay.container_image_pusher.QuayClient")
def test_push_container_items_v1_item(
    mock_quay_client,
    mock_copy_multiarch,
    mock_copy_src,
    target_settings,
    container_multiarch_push_item,
):
    mock_get_manifest = mock.MagicMock()

    mock_get_manifest.side_effect = exceptions.ManifestTypeError("no manifest list")
    mock_quay_client.return_value.get_manifest = mock_get_manifest

    pusher = container_image_pusher.ContainerImagePusher(
        [container_multiarch_push_item], target_settings
    )
    pusher.push_container_images()


@mock.patch("pubtools._quay.container_image_pusher.ContainerImagePusher.copy_v1_push_item")
@mock.patch("pubtools._quay.container_image_pusher.ContainerImagePusher.copy_multiarch_push_item")
@mock.patch("pubtools._quay.container_image_pusher.QuayClient")
def test_push_container_items_single_arch(
    mock_quay_client,
    mock_copy_multiarch,
    mock_copy_src,
    target_settings_allow_v1_containers_false,
    container_multiarch_push_item,
):
    mock_get_manifest = mock.MagicMock()

    mock_get_manifest.side_effect = exceptions.ManifestTypeError("no manifest list")
    mock_quay_client.return_value.get_manifest = mock_get_manifest

    pusher = container_image_pusher.ContainerImagePusher(
        [container_multiarch_push_item], target_settings_allow_v1_containers_false
    )
    pusher.push_container_images()


@mock.patch("pubtools._quay.container_image_pusher.ContainerImagePusher.copy_source_push_item")
@mock.patch("pubtools._quay.container_image_pusher.ContainerImagePusher.copy_multiarch_push_item")
@mock.patch("pubtools._quay.container_image_pusher.QuayClient")
def test_push_container_items_multiarch_item(
    mock_quay_client,
    mock_copy_multiarch,
    mock_copy_src,
    target_settings,
    container_multiarch_push_item,
):
    mock_get_manifest = mock.MagicMock()

    mock_get_manifest.return_value = {"some-manifest": "manifest-list"}
    mock_quay_client.return_value.get_manifest = mock_get_manifest

    pusher = container_image_pusher.ContainerImagePusher(
        [container_multiarch_push_item], target_settings
    )
    pusher.push_container_images()

    mock_copy_src.assert_not_called()
    mock_copy_multiarch.assert_called_once()


@mock.patch("pubtools._quay.container_image_pusher.ContainerImagePusher.copy_source_push_item")
@mock.patch("pubtools._quay.container_image_pusher.ContainerImagePusher.copy_multiarch_push_item")
@mock.patch("pubtools._quay.container_image_pusher.QuayClient")
def test_push_container_items_multiple_items(
    mock_quay_client,
    mock_copy_multiarch,
    mock_copy_src,
    target_settings,
    container_multiarch_push_item,
):
    mock_get_manifest = mock.MagicMock()

    mock_get_manifest.return_value = {"some-manifest": "manifest-list"}
    mock_quay_client.return_value.get_manifest = mock_get_manifest

    push_items = [copy.deepcopy(container_multiarch_push_item) for i in range(10)]

    pusher = container_image_pusher.ContainerImagePusher(push_items, target_settings)
    pusher.push_container_images()

    assert mock_copy_multiarch.call_count == 10



import logging
import mock
import pytest
import requests_mock
import requests

from pubtools._quay import exceptions
from pubtools._quay import quay_client
from pubtools._quay import operator_pusher
from .utils.misc import sort_dictionary_sortable_values, compare_logs, IIBRes

# flake8: noqa: E501


def test_init(target_settings, operator_push_item_ok):
    pusher = operator_pusher.OperatorPusher([operator_push_item_ok], "3", target_settings)

    assert pusher.push_items == [operator_push_item_ok]
    assert pusher.task_id == "3"
    assert pusher.target_settings == target_settings
    assert pusher.quay_host == "quay.io"


def test_get_immutable_tag_vr(target_settings, operator_push_item_vr):
    pusher = operator_pusher.OperatorPusher([operator_push_item_vr], "3", target_settings)

    tag = pusher._get_immutable_tag(operator_push_item_vr)
    assert tag == "1.0"


def test_get_immutable_tag_no_vr(target_settings, operator_push_item_no_vr):
    pusher = operator_pusher.OperatorPusher([operator_push_item_no_vr], "3", target_settings)

    tag = pusher._get_immutable_tag(operator_push_item_no_vr)
    assert tag == "1.0000000"


def test_public_bundle_ref(target_settings, operator_push_item_no_vr):
    pusher = operator_pusher.OperatorPusher([operator_push_item_no_vr], "3", target_settings)

    ref = pusher.public_bundle_ref(operator_push_item_no_vr)
    assert ref == "some-registry1.com/repo1:1.0000000"


@mock.patch("pubtools._quay.operator_pusher.run_entrypoint")
def test_pyxis_get_ocp_versions(
    mock_run_entrypoint, target_settings, operator_push_item_ok, fake_cert_key_paths
):
    pusher = operator_pusher.OperatorPusher([operator_push_item_ok], "3", target_settings)

    mock_run_entrypoint.return_value = [{"ocp_version": "4.5"}, {"ocp_version": "4.6"}]
    versions = pusher.pyxis_get_ocp_versions(operator_push_item_ok)

    mock_run_entrypoint.assert_called_once_with(
        ("pubtools-pyxis", "console_scripts", "pubtools-pyxis-get-operator-indices"),
        "pubtools-pyxis-get-operator-indices",
        [
            "--pyxis-server",
            "pyxis-url.com",
            "--pyxis-ssl-crtfile",
            "/path/to/file.crt",
            "--pyxis-ssl-keyfile",
            "/path/to/file.key",
            "--organization",
            "redhat-operators",
            "--ocp-versions-range",
            "v4.5",
        ],
        {},
    )
    assert versions == ["v4.5", "v4.6"]


@mock.patch("pubtools._quay.operator_pusher.run_entrypoint")
def test_pyxis_get_ocp_versions_no_data(
    mock_run_entrypoint, target_settings, operator_push_item_ok, fake_cert_key_paths
):
    pusher = operator_pusher.OperatorPusher([operator_push_item_ok], "3", target_settings)

    mock_run_entrypoint.return_value = []
    with pytest.raises(ValueError, match="Pyxis has returned no OCP.*"):
        versions = pusher.pyxis_get_ocp_versions(operator_push_item_ok)


@mock.patch("pubtools._quay.operator_pusher.run_entrypoint")
def test_pyxis_generate_mapping(
    mock_run_entrypoint,
    target_settings,
    operator_push_item_ok,
    operator_push_item_different_version,
    fake_cert_key_paths,
):
    mock_run_entrypoint.side_effect = [
        [{"ocp_version": "4.5"}, {"ocp_version": "4.6"}, {"ocp_version": "4.7"}],
        [{"ocp_version": "4.7"}],
    ]
    pusher = operator_pusher.OperatorPusher(
        [operator_push_item_ok, operator_push_item_different_version], "3", target_settings
    )

    mapping = pusher.version_items_mapping
    assert mock_run_entrypoint.call_count == 2
    assert len(mapping["v4.5"]) == 1
    assert len(mapping["v4.6"]) == 1
    assert len(mapping["v4.7"]) == 2


def test_get_deprecation_list(target_settings, operator_push_item_ok):
    pusher = operator_pusher.OperatorPusher([operator_push_item_ok], "3", target_settings)
    with open("tests/test_data/deprecation_list_data.yaml", "r") as f:
        deprecate_data = f.read()

    with requests_mock.Mocker() as m:
        m.get("https://git-server.com/4_7.yml/raw?ref=master", text=deprecate_data)
        deprecation_list = pusher.get_deprecation_list("4.7")

    assert deprecation_list == [
        "some-registry1.com/bundle/path@sha256:a1a1a1",
        "some-registry1.com/bundle/path@sha256:b2b2b2",
    ]


def test_get_deprecation_list_server_error(target_settings, operator_push_item_ok, caplog):
    pusher = operator_pusher.OperatorPusher([operator_push_item_ok], "3", target_settings)

    with requests_mock.Mocker() as m:
        m.get("https://git-server.com/4_7.yml/raw?ref=master", status_code=500)
        with pytest.raises(requests.exceptions.HTTPError, match=".*500.*"):
            deprecation_list = pusher.get_deprecation_list("4.7")


def test_get_deprecation_list_invalid_data(target_settings, operator_push_item_ok):
    pusher = operator_pusher.OperatorPusher([operator_push_item_ok], "3", target_settings)

    with requests_mock.Mocker() as m:
        m.get("https://git-server.com/4_7.yml/raw?ref=master", text="{some-invalid-data}")
        with pytest.raises(TypeError, match=".*not iterable.*"):
            deprecation_list = pusher.get_deprecation_list("4.7")


def test_get_deprecation_list_no_url(target_settings, operator_push_item_ok):
    target_settings["iib_deprecation_list_url"] = None
    pusher = operator_pusher.OperatorPusher([operator_push_item_ok], "3", target_settings)

    deprecation_list = pusher.get_deprecation_list("4.7")

    assert deprecation_list == None


@mock.patch("pubtools._quay.operator_pusher.run_entrypoint")
def test_iib_add_bundles_str_deprecation_list_build_timeout(
    mock_run_entrypoint, target_settings, operator_push_item_ok
):
    mock_run_entrypoint.return_value = "some-data"
    target_settings["iib_build_timeout"] = 30
    pusher = operator_pusher.OperatorPusher([operator_push_item_ok], "3", target_settings)
    result = pusher.iib_add_bundles(
        ["bundle1", "bundle2"],
        ["arch1", "arch2"],
        "registry.com/rh-osbs/iib-pub-pending:v4.5",
        "bundle3,bundle4",
        ["tag1", "tag2"],
        pusher.target_settings,
    )

    assert result == "some-data"
    mock_run_entrypoint.assert_called_once_with(
        ("pubtools-iib", "console_scripts", "pubtools-iib-add-bundles"),
        "pubtools-iib-add-bundles",
        [
            "--iib-server",
            "iib-server.com",
            "--iib-krb-principal",
            "some-principal@REDHAT.COM",
            "--overwrite-from-index",
            "--iib-krb-ktfile",
            "/etc/pub/some.keytab",
            "--build-timeout",
            "30",
            "--index-image",
            "registry.com/rh-osbs/iib-pub-pending:v4.5",
            "--bundle",
            "bundle1",
            "--bundle",
            "bundle2",
            "--arch",
            "arch1",
            "--arch",
            "arch2",
            "--deprecation-list",
            "bundle3,bundle4",
            "--build-tag",
            "tag1",
            "--build-tag",
            "tag2",
        ],
        {"OVERWRITE_FROM_INDEX_TOKEN": "some-user:some-pass"},
    )


@mock.patch("pubtools._quay.operator_pusher.run_entrypoint")
def test_iib_add_bundles_error(mock_run_entrypoint, target_settings, operator_push_item_ok):
    mock_run_entrypoint.side_effect = SystemExit(1)
    pusher = operator_pusher.OperatorPusher([operator_push_item_ok], "3", target_settings)
    result = pusher.iib_add_bundles(
        ["bundle1", "bundle2"],
        ["arch1", "arch2"],
        "registry.com/rh-osbs/iib-pub-pending:v4.5",
        "bundle3,bundle4",
        ["tag1", "tag2"],
        pusher.target_settings,
    )

    assert result == False


@mock.patch("pubtools._quay.operator_pusher.run_entrypoint")
def test_iib_add_bundles_list_deprecation_list(
    mock_run_entrypoint, target_settings, operator_push_item_ok
):
    mock_run_entrypoint.return_value = "some-data"
    target_settings["check_related_images"] = True
    pusher = operator_pusher.OperatorPusher([operator_push_item_ok], "3", target_settings)
    result = pusher.iib_add_bundles(
        ["bundle1", "bundle2"],
        ["arch1", "arch2"],
        "registry.com/rh-osbs/iib-pub-pending:v4.5",
        ["bundle3", "bundle4"],
        None,
        pusher.target_settings,
    )

    assert result == "some-data"
    mock_run_entrypoint.assert_called_once_with(
        ("pubtools-iib", "console_scripts", "pubtools-iib-add-bundles"),
        "pubtools-iib-add-bundles",
        [
            "--iib-server",
            "iib-server.com",
            "--iib-krb-principal",
            "some-principal@REDHAT.COM",
            "--overwrite-from-index",
            "--iib-krb-ktfile",
            "/etc/pub/some.keytab",
            "--check-related-images",
            "--index-image",
            "registry.com/rh-osbs/iib-pub-pending:v4.5",
            "--bundle",
            "bundle1",
            "--bundle",
            "bundle2",
            "--arch",
            "arch1",
            "--arch",
            "arch2",
            "--deprecation-list",
            "bundle3,bundle4",
        ],
        {"OVERWRITE_FROM_INDEX_TOKEN": "some-user:some-pass"},
    )


@mock.patch("pubtools._quay.operator_pusher.run_entrypoint")
def test_iib_remove_operators(mock_run_entrypoint, target_settings, operator_push_item_ok):
    mock_run_entrypoint.return_value = "some-data"
    pusher = operator_pusher.OperatorPusher([operator_push_item_ok], "3", target_settings)
    result = pusher.iib_remove_operators(
        ["operator1", "operator2"],
        ["arch1", "arch2"],
        "registry.com/rh-osbs/iib-pub-pending:v4.5",
        ["tag1", "tag2"],
        pusher.target_settings,
    )

    assert result == "some-data"
    mock_run_entrypoint.assert_called_once_with(
        ("pubtools-iib", "console_scripts", "pubtools-iib-remove-operators"),
        "pubtools-iib-remove-operators",
        [
            "--iib-server",
            "iib-server.com",
            "--iib-krb-principal",
            "some-principal@REDHAT.COM",
            "--overwrite-from-index",
            "--iib-krb-ktfile",
            "/etc/pub/some.keytab",
            "--index-image",
            "registry.com/rh-osbs/iib-pub-pending:v4.5",
            "--operator",
            "operator1",
            "--operator",
            "operator2",
            "--arch",
            "arch1",
            "--arch",
            "arch2",
            "--build-tag",
            "tag1",
            "--build-tag",
            "tag2",
        ],
        {"OVERWRITE_FROM_INDEX_TOKEN": "some-user:some-pass"},
    )


@mock.patch("pubtools._quay.operator_pusher.run_entrypoint")
def test_iib_add_deprecations(mock_run_entrypoint, target_settings, operator_push_item_ok):
    mock_run_entrypoint.return_value = "some-data"
    pusher = operator_pusher.OperatorPusher([operator_push_item_ok], "3", target_settings)
    result = pusher.iib_add_deprecations(
        "registry.com/rh-osbs/iib-pub-pending:v4.5",
        '{"a": "b"}',
        "operator1",
        ["tag1", "tag2"],
        pusher.target_settings,
    )

    assert result == "some-data"
    mock_run_entrypoint.assert_called_once_with(
        ("pubtools-iib", "console_scripts", "pubtools-iib-add-deprecations"),
        "pubtools-iib-add-deprecations",
        [
            "--iib-server",
            "iib-server.com",
            "--iib-krb-principal",
            "some-principal@REDHAT.COM",
            "--overwrite-from-index",
            "--iib-krb-ktfile",
            "/etc/pub/some.keytab",
            "--index-image",
            "registry.com/rh-osbs/iib-pub-pending:v4.5",
            "--deprecation-schema",
            '{"a": "b"}',
            "--operator-package",
            "operator1",
            "--build-tag",
            "tag1",
            "--build-tag",
            "tag2",
        ],
        {"OVERWRITE_FROM_INDEX_TOKEN": "some-user:some-pass"},
    )


@mock.patch("pubtools._quay.operator_pusher.ContainerImagePusher.run_tag_images")
@mock.patch("pubtools._quay.operator_pusher.OperatorPusher.iib_add_bundles")
@mock.patch("pubtools._quay.operator_pusher.run_entrypoint")
@mock.patch("pubtools._quay.operator_pusher.OperatorPusher.get_deprecation_list")
@mock.patch("pubtools._quay.operator_pusher.pyxis_get_repo_metadata")
def test_push_operators(
    mock_get_repo_metadata,
    mock_get_deprecation_list,
    mock_run_entrypoint,
    mock_add_bundles,
    mock_run_tag_images,
    target_settings,
    operator_push_item_ok,
    operator_push_item_different_version,
    fake_cert_key_paths,
):
    mock_get_repo_metadata.side_effect = [
        {"fbc_opt_in": False},
        {"fbc_opt_in": False},
        {"fbc_opt_in": False},
    ]

    mock_get_deprecation_list.side_effect = [["bundle1", "bundle2"], ["bundle3"], []]

    mock_run_entrypoint.side_effect = [
        [{"ocp_version": "4.5"}, {"ocp_version": "4.6"}, {"ocp_version": "4.7"}],
        [{"ocp_version": "4.7"}],
    ]
    iib_results = [
        IIBRes(
            "some-registry.com/ns/index-image:5",
            "some-registry.com/ns/iib@sha256:a1a1",
            ["v4.5-3"],
        ),
        IIBRes(
            "some-registry.com/ns/index-image:6",
            "some-registry.com/ns/iib@sha256:b2b2",
            ["v4.6-3"],
        ),
        IIBRes(
            "some-registry.com/ns/index-image:7",
            "some-registry.com/ns/iib@sha256:c3c3",
            ["v4.7-3"],
        ),
    ]
    mock_add_bundles.side_effect = iib_results
    pusher = operator_pusher.OperatorPusher(
        [operator_push_item_ok, operator_push_item_different_version], "3", target_settings
    )

    results = pusher.build_index_images()

    assert mock_get_deprecation_list.call_count == 3
    assert mock_get_deprecation_list.call_args_list[0] == mock.call("v4.5")
    assert mock_get_deprecation_list.call_args_list[1] == mock.call("v4.6")
    assert mock_get_deprecation_list.call_args_list[2] == mock.call("v4.7")

    assert results == {
        "v4.5-v4.5": {
            "iib_result": iib_results[0],
            "signing_keys": ["some-key"],
            "destination_tags": ["v4.5"],
        },
        "v4.6-v4.6": {
            "iib_result": iib_results[1],
            "signing_keys": ["some-key"],
            "destination_tags": ["v4.6"],
        },
        "v4.7-v4.7": {
            "iib_result": iib_results[2],
            "signing_keys": ["some-key"],
            "destination_tags": ["v4.7"],
        },
    }
    assert mock_add_bundles.call_count == 3
    assert mock_add_bundles.call_args_list[0] == mock.call(
        bundles=["some-registry1.com/repo:1.0"],
        index_image="registry.com/rh-osbs/iib-pub-pending:v4.5",
        deprecation_list=["bundle1", "bundle2"],
        build_tags=["v4.5-3"],
        target_settings=target_settings,
    )
    assert mock_add_bundles.call_args_list[1] == mock.call(
        bundles=["some-registry1.com/repo:1.0"],
        index_image="registry.com/rh-osbs/iib-pub-pending:v4.6",
        deprecation_list=["bundle3"],
        build_tags=["v4.6-3"],
        target_settings=target_settings,
    )
    assert mock_add_bundles.call_args_list[2] == mock.call(
        bundles=["some-registry1.com/repo:1.0", "some-registry1.com/repo2:5.0.0"],
        index_image="registry.com/rh-osbs/iib-pub-pending:v4.7",
        deprecation_list=[],
        build_tags=["v4.7-3"],
        target_settings=target_settings,
    )

    pusher.push_index_images(results)

    assert mock_run_tag_images.call_count == 3
    mock_run_tag_images.assert_has_calls(
        [
            mock.call(
                "some-registry.com/ns/index-image:5",
                ["quay.io/some-namespace/operators----index-image:v4.5"],
                True,
                target_settings,
            )
        ]
    )
    mock_run_tag_images.assert_has_calls(
        [
            mock.call(
                "some-registry.com/ns/index-image:6",
                ["quay.io/some-namespace/operators----index-image:v4.6"],
                True,
                target_settings,
            )
        ]
    )
    mock_run_tag_images.assert_has_calls(
        [
            mock.call(
                "some-registry.com/ns/index-image:7",
                ["quay.io/some-namespace/operators----index-image:v4.7"],
                True,
                target_settings,
            )
        ]
    )


@mock.patch("pubtools._quay.operator_pusher.ContainerImagePusher.run_tag_images")
@mock.patch("pubtools._quay.operator_pusher.OperatorPusher.iib_add_bundles")
@mock.patch("pubtools._quay.operator_pusher.run_entrypoint")
@mock.patch("pubtools._quay.operator_pusher.OperatorPusher.get_deprecation_list")
@mock.patch("pubtools._quay.operator_pusher.pyxis_get_repo_metadata")
def test_push_operators_extra_ns(
    mock_get_repo_metadata,
    mock_get_deprecation_list,
    mock_run_entrypoint,
    mock_add_bundles,
    mock_run_tag_images,
    target_settings,
    operator_push_item_ok,
    operator_push_item_different_version,
    fake_cert_key_paths,
):
    """Test if extra namespace is used in tagging when is set in target settings"""

    mock_get_repo_metadata.side_effect = [
        {"fbc_opt_in": False},
        {"fbc_opt_in": False},
        {"fbc_opt_in": False},
    ]

    target_settings["quay_operator_namespace"] = "quay-operator-ns"
    mock_get_deprecation_list.side_effect = [["bundle1", "bundle2"], ["bundle3"], []]

    mock_run_entrypoint.side_effect = [
        [{"ocp_version": "4.5"}, {"ocp_version": "4.6"}, {"ocp_version": "4.7"}],
        [{"ocp_version": "4.7"}],
    ]
    iib_results = [
        IIBRes(
            "some-registry.com/ns/index-image:5",
            "some-registry.com/ns/iib@sha256:a1a1",
            ["v4.5-3"],
        ),
        IIBRes(
            "some-registry.com/ns/index-image:6",
            "some-registry.com/ns/iib@sha256:b2b2",
            ["v4.6-3"],
        ),
        IIBRes(
            "some-registry.com/ns/index-image:7",
            "some-registry.com/ns/iib@sha256:c3c3",
            ["v4.7-3"],
        ),
    ]
    mock_add_bundles.side_effect = iib_results
    pusher = operator_pusher.OperatorPusher(
        [operator_push_item_ok, operator_push_item_different_version], "3", target_settings
    )

    results = pusher.build_index_images()

    pusher.push_index_images(results)

    assert mock_run_tag_images.call_count == 3
    # TODO: Unstable tests, the order has to be fixed before these checks can be enabled
    # mock_run_tag_images.assert_has_calls(
    #     [
    #         mock.call(
    #             "some-registry.com/ns/index-image:5",
    #             ["quay.io/quay-operator-ns/operators----index-image:v4.5"],
    #             True,
    #             target_settings,
    #         )
    #     ]
    # )
    # mock_run_tag_images.assert_has_calls(
    #     [
    #         mock.call(
    #             "some-registry.com/ns/index-image:6",
    #             ["quay.io/quay-operator-ns/operators----index-image:v4.6"],
    #             True,
    #             target_settings,
    #         )
    #     ]
    # )
    # mock_run_tag_images.assert_has_calls(
    #     [
    #         mock.call(
    #             "some-registry.com/ns/index-image:7",
    #             ["quay.io/quay-operator-ns/operators----index-image:v4.7"],
    #             True,
    #             target_settings,
    #         )
    #     ]
    # )


class Mock_iib_add_bundles:
    def __init__(self, responses):
        self.responses = responses

    def __call__(
        self,
        bundles=None,
        archs=None,
        index_image=None,
        deprecation_list=None,
        build_tags=None,
        target_settings={},
    ):
        return self.responses[index_image]


@mock.patch("pubtools._quay.operator_pusher.ContainerImagePusher.run_tag_images")
@mock.patch("pubtools._quay.operator_pusher.OperatorPusher.iib_add_bundles")
@mock.patch("pubtools._quay.operator_pusher.run_entrypoint")
@mock.patch("pubtools._quay.operator_pusher.OperatorPusher.get_deprecation_list")
@mock.patch("pubtools._quay.operator_pusher.pyxis_get_repo_metadata")
def test_push_operators_not_all_successful(
    mock_get_repo_metadata,
    mock_get_deprecation_list,
    mock_run_entrypoint,
    mock_add_bundles,
    mock_run_tag_images,
    target_settings,
    operator_push_item_ok,
    operator_push_item_different_version,
    fake_cert_key_paths,
):
    mock_get_repo_metadata.side_effect = [
        {"fbc_opt_in": False},
        {"fbc_opt_in": False},
        {"fbc_opt_in": False},
    ]

    mock_get_deprecation_list.side_effect = [["bundle1", "bundle2"], ["bundle3"], []]

    mock_run_entrypoint.side_effect = [
        [{"ocp_version": "4.5"}, {"ocp_version": "4.6"}, {"ocp_version": "4.7"}],
        [{"ocp_version": "4.7"}],
    ]
    iib_results = [
        IIBRes(
            "some-registry.com/ns/index-image:5",
            "some-registry.com/ns/iib@sha256:a1a1",
            ["v4.5-3"],
        ),
        None,
        IIBRes(
            "some-registry.com/ns/index-image:7",
            "some-registry.com/ns/iib@sha256:c3c3",
            ["v4.7-3"],
        ),
    ]
    mock_add_bundles.side_effect = Mock_iib_add_bundles(
        {
            "registry.com/rh-osbs/iib-pub-pending:v4.5": iib_results[0],
            "registry.com/rh-osbs/iib-pub-pending:v4.6": None,
            "registry.com/rh-osbs/iib-pub-pending:v4.7": iib_results[2],
        }
    )

    iib_results
    pusher = operator_pusher.OperatorPusher(
        [operator_push_item_ok, operator_push_item_different_version], "3", target_settings
    )

    results = pusher.build_index_images()

    assert mock_get_deprecation_list.call_count == 3
    assert mock_get_deprecation_list.call_args_list[0] == mock.call("v4.5")
    assert mock_get_deprecation_list.call_args_list[1] == mock.call("v4.6")
    assert mock_get_deprecation_list.call_args_list[2] == mock.call("v4.7")

    assert results == {
        "v4.5-v4.5": {
            "iib_result": iib_results[0],
            "signing_keys": ["some-key"],
            "destination_tags": ["v4.5"],
        },
        "v4.6-v4.6": {
            "iib_result": None,
            "signing_keys": ["some-key"],
            "destination_tags": ["v4.6"],
        },
        "v4.7-v4.7": {
            "iib_result": iib_results[2],
            "signing_keys": ["some-key"],
            "destination_tags": ["v4.7"],
        },
    }
    assert mock_add_bundles.call_count == 3
    assert (
        mock.call(
            bundles=["some-registry1.com/repo:1.0"],
            index_image="registry.com/rh-osbs/iib-pub-pending:v4.5",
            deprecation_list=["bundle1", "bundle2"],
            build_tags=["v4.5-3"],
            target_settings=target_settings,
        )
        in mock_add_bundles.call_args_list
    )
    assert (
        mock.call(
            bundles=["some-registry1.com/repo:1.0"],
            index_image="registry.com/rh-osbs/iib-pub-pending:v4.6",
            deprecation_list=["bundle3"],
            build_tags=["v4.6-3"],
            target_settings=target_settings,
        )
        in mock_add_bundles.call_args_list
    )
    assert (
        mock.call(
            bundles=["some-registry1.com/repo:1.0", "some-registry1.com/repo2:5.0.0"],
            index_image="registry.com/rh-osbs/iib-pub-pending:v4.7",
            deprecation_list=[],
            build_tags=["v4.7-3"],
            target_settings=target_settings,
        )
        in mock_add_bundles.call_args_list
    )

    pusher.push_index_images(results)

    assert mock_run_tag_images.call_count == 2
    mock_run_tag_images.assert_has_calls(
        [
            mock.call(
                "some-registry.com/ns/index-image:5",
                ["quay.io/some-namespace/operators----index-image:v4.5"],
                True,
                target_settings,
            )
        ]
    )
    mock_run_tag_images.assert_has_calls(
        [
            mock.call(
                "some-registry.com/ns/index-image:7",
                ["quay.io/some-namespace/operators----index-image:v4.7"],
                True,
                target_settings,
            )
        ]
    )


@mock.patch("pubtools._quay.operator_pusher.ContainerImagePusher.run_tag_images")
@mock.patch("pubtools._quay.operator_pusher.OperatorPusher.iib_add_bundles")
@mock.patch("pubtools._quay.operator_pusher.run_entrypoint")
@mock.patch("pubtools._quay.operator_pusher.OperatorPusher.get_deprecation_list")
@mock.patch("pubtools._quay.operator_pusher.pyxis_get_repo_metadata")
def test_push_operators_hotfix(
    mock_get_repo_metadata,
    mock_get_deprecation_list,
    mock_run_entrypoint,
    mock_add_bundles,
    mock_run_tag_images,
    target_settings,
    operator_push_item_hotfix,
    fake_cert_key_paths,
):
    mock_get_repo_metadata.side_effect = [
        {"fbc_opt_in": False},
        {"fbc_opt_in": False},
        {"fbc_opt_in": False},
    ]

    mock_get_deprecation_list.side_effect = [["bundle1", "bundle2"], ["bundle3"]]

    mock_run_entrypoint.side_effect = [
        [{"ocp_version": "4.5"}, {"ocp_version": "4.6"}],
    ]
    iib_results = [
        IIBRes(
            "some-registry.com/index/image:5",
            "some-registry.com/index/image@sha256:a1a1",
            ["v4.5-test-hotfix-RHBA-1234-4567"],
        ),
        IIBRes(
            "some-registry.com/index/image:6",
            "some-registry.com/index/image@sha256:b2b2",
            ["v4.6-test-hotfix-RHBA-1234-4567"],
        ),
    ]
    mock_add_bundles.side_effect = iib_results
    pusher = operator_pusher.OperatorPusher([operator_push_item_hotfix], "3", target_settings)

    results = pusher.build_index_images()

    assert mock_get_deprecation_list.call_count == 2
    assert mock_get_deprecation_list.call_args_list[0] == mock.call("v4.5")
    assert mock_get_deprecation_list.call_args_list[1] == mock.call("v4.6")

    assert results == {
        "v4.5-RHBA-1234:4567": {
            "iib_result": iib_results[0],
            "signing_keys": ["some-key"],
            "destination_tags": ["v4.5-test-hotfix-1234-4567"],
        },
        "v4.6-RHBA-1234:4567": {
            "iib_result": iib_results[1],
            "signing_keys": ["some-key"],
            "destination_tags": ["v4.6-test-hotfix-1234-4567"],
        },
    }
    expected_target_settings = target_settings.copy()
    expected_target_settings["iib_overwrite_from_index"] = False
    expected_target_settings["iib_overwrite_from_index_token"] = ""
    assert mock_add_bundles.call_count == 2
    assert mock_add_bundles.call_args_list[0] == mock.call(
        bundles=["some-registry1.com/repo:1.0"],
        index_image="registry.com/rh-osbs/iib-pub-pending:v4.5",
        deprecation_list=["bundle1", "bundle2"],
        build_tags=["v4.5-3"],
        target_settings=expected_target_settings,
    )
    assert mock_add_bundles.call_args_list[1] == mock.call(
        bundles=["some-registry1.com/repo:1.0"],
        index_image="registry.com/rh-osbs/iib-pub-pending:v4.6",
        deprecation_list=["bundle3"],
        build_tags=["v4.6-3"],
        target_settings=expected_target_settings,
    )

    pusher.push_index_images(results)

    assert mock_run_tag_images.call_count == 2
    mock_run_tag_images.assert_has_calls(
        [
            mock.call(
                "some-registry.com/index/image:5",
                ["quay.io/some-namespace/operators----index-image:v4.5-test-hotfix-1234-4567"],
                True,
                target_settings,
            )
        ]
    )
    mock_run_tag_images.assert_has_calls(
        [
            mock.call(
                "some-registry.com/index/image:6",
                ["quay.io/some-namespace/operators----index-image:v4.6-test-hotfix-1234-4567"],
                True,
                target_settings,
            )
        ]
    )


@mock.patch("pubtools._quay.operator_pusher.ContainerImagePusher.run_tag_images")
@mock.patch("pubtools._quay.operator_pusher.OperatorPusher.iib_add_bundles")
@mock.patch("pubtools._quay.operator_pusher.run_entrypoint")
@mock.patch("pubtools._quay.operator_pusher.OperatorPusher.get_deprecation_list")
@mock.patch("pubtools._quay.operator_pusher.pyxis_get_repo_metadata")
def test_push_operators_prerelease(
    mock_get_repo_metadata,
    mock_get_deprecation_list,
    mock_run_entrypoint,
    mock_add_bundles,
    mock_run_tag_images,
    target_settings,
    operator_push_item_pre_release,
    operator_push_item_pre_release2,
    fake_cert_key_paths,
):
    mock_get_repo_metadata.side_effect = [
        {"fbc_opt_in": False},
        {"fbc_opt_in": False},
        {"fbc_opt_in": False},
    ]

    mock_get_deprecation_list.side_effect = [["bundle1", "bundle2"], ["bundle3"]]

    mock_run_entrypoint.side_effect = [
        [{"ocp_version": "4.5"}, {"ocp_version": "4.6"}],
    ]
    iib_results = [
        IIBRes(
            "some-registry.com/index/image:v4.5",
            "some-registry.com/index/image@sha256:a1a1",
            ["v4.5-prerelease-operator-name-1234-4567"],
        ),
        IIBRes(
            "some-registry.com/index/image:v4.5",
            "some-registry.com/index/image@sha256:a1a1",
            ["v4.5-prerelease-operator-name2-1234-4567"],
        ),
        IIBRes(
            "some-registry.com/index/image:v4.6",
            "some-registry.com/index/image@sha256:b2b2",
            ["v4.6-prerelease-operator-name-1234-4567"],
        ),
        IIBRes(
            "some-registry.com/index/image:v4.6",
            "some-registry.com/index/image@sha256:b2b2",
            ["v4.6-prerelease-operator-name2-1234-4567"],
        ),
    ]
    mock_add_bundles.side_effect = iib_results
    pusher = operator_pusher.OperatorPusher(
        [operator_push_item_pre_release, operator_push_item_pre_release2], "3", target_settings
    )

    results = pusher.build_index_images()

    assert mock_get_deprecation_list.call_count == 2
    assert mock_get_deprecation_list.call_args_list[0] == mock.call("v4.5")
    assert mock_get_deprecation_list.call_args_list[1] == mock.call("v4.6")

    assert results == {
        "v4.5-RHBA-1234:4567-operator-name": {
            "iib_result": iib_results[0],
            "signing_keys": ["some-key"],
            "destination_tags": [
                "v4.5-prerelease-operator-name-1234-4567",
                "v4.5-prerelease-operator-name",
            ],
        },
        "v4.5-RHBA-1234:4567-operator-name2": {
            "iib_result": iib_results[1],
            "signing_keys": ["some-key"],
            "destination_tags": [
                "v4.5-prerelease-operator-name2-1234-4567",
                "v4.5-prerelease-operator-name2",
            ],
        },
        "v4.6-RHBA-1234:4567-operator-name": {
            "iib_result": iib_results[2],
            "signing_keys": ["some-key"],
            "destination_tags": [
                "v4.6-prerelease-operator-name-1234-4567",
                "v4.6-prerelease-operator-name",
            ],
        },
        "v4.6-RHBA-1234:4567-operator-name2": {
            "iib_result": iib_results[3],
            "signing_keys": ["some-key"],
            "destination_tags": [
                "v4.6-prerelease-operator-name2-1234-4567",
                "v4.6-prerelease-operator-name2",
            ],
        },
    }
    expected_target_settings = target_settings.copy()
    expected_target_settings["iib_overwrite_from_index"] = False
    expected_target_settings["iib_overwrite_from_index_token"] = ""
    assert mock_add_bundles.call_count == 4
    assert mock_add_bundles.call_args_list[0] == mock.call(
        bundles=["some-registry1.com/repo:1.0"],
        index_image="registry.com/rh-osbs/iib-pub-pending:v4.5",
        deprecation_list=["bundle1", "bundle2"],
        build_tags=["v4.5-3-RHBA-1234-4567-operator-name"],
        target_settings=expected_target_settings,
    )
    assert mock_add_bundles.call_args_list[1] == mock.call(
        bundles=["some-registry1.com/repo:1.0"],
        index_image="registry.com/rh-osbs/iib-pub-pending:v4.5",
        deprecation_list=["bundle1", "bundle2"],
        build_tags=["v4.5-3-RHBA-1234-4567-operator-name2"],
        target_settings=expected_target_settings,
    )
    assert mock_add_bundles.call_args_list[2] == mock.call(
        bundles=["some-registry1.com/repo:1.0"],
        index_image="registry.com/rh-osbs/iib-pub-pending:v4.6",
        deprecation_list=["bundle3"],
        build_tags=["v4.6-3-RHBA-1234-4567-operator-name"],
        target_settings=expected_target_settings,
    )
    assert mock_add_bundles.call_args_list[3] == mock.call(
        bundles=["some-registry1.com/repo:1.0"],
        index_image="registry.com/rh-osbs/iib-pub-pending:v4.6",
        deprecation_list=["bundle3"],
        build_tags=["v4.6-3-RHBA-1234-4567-operator-name2"],
        target_settings=expected_target_settings,
    )

    pusher.push_index_images(results)

    assert mock_run_tag_images.call_count == 4
    mock_run_tag_images.assert_has_calls(
        [
            mock.call(
                "some-registry.com/index/image:v4.6",
                [
                    "quay.io/some-namespace/operators----index-image:v4.6-prerelease-operator-name-1234-4567",
                    "quay.io/some-namespace/operators----index-image:v4.6-prerelease-operator-name",
                ],
                True,
                target_settings,
            )
        ]
    )
    mock_run_tag_images.assert_has_calls(
        [
            mock.call(
                "some-registry.com/index/image:v4.5",
                [
                    "quay.io/some-namespace/operators----index-image:v4.5-prerelease-operator-name-1234-4567",
                    "quay.io/some-namespace/operators----index-image:v4.5-prerelease-operator-name",
                ],
                True,
                target_settings,
            )
        ]
    )
    mock_run_tag_images.assert_has_calls(
        [
            mock.call(
                "some-registry.com/index/image:v4.6",
                [
                    "quay.io/some-namespace/operators----index-image:v4.6-prerelease-operator-name2-1234-4567",
                    "quay.io/some-namespace/operators----index-image:v4.6-prerelease-operator-name2",
                ],
                True,
                target_settings,
            )
        ]
    )
    mock_run_tag_images.assert_has_calls(
        [
            mock.call(
                "some-registry.com/index/image:v4.5",
                [
                    "quay.io/some-namespace/operators----index-image:v4.5-prerelease-operator-name2-1234-4567",
                    "quay.io/some-namespace/operators----index-image:v4.5-prerelease-operator-name2",
                ],
                True,
                target_settings,
            ),
        ]
    )


@mock.patch("pubtools._quay.operator_pusher.ContainerImagePusher.run_tag_images")
@mock.patch("pubtools._quay.operator_pusher.OperatorPusher.iib_add_bundles")
@mock.patch("pubtools._quay.operator_pusher.run_entrypoint")
@mock.patch("pubtools._quay.operator_pusher.OperatorPusher.get_deprecation_list")
@mock.patch("pubtools._quay.operator_pusher.pyxis_get_repo_metadata")
def test_push_operators_hotfix_invalid_origin(
    mock_get_repo_metadata,
    mock_get_deprecation_list,
    mock_run_entrypoint,
    mock_add_bundles,
    mock_run_tag_images,
    target_settings,
    operator_push_item_hotfix_invalid_origin,
    fake_cert_key_paths,
):
    mock_get_repo_metadata.side_effect = [
        {"fbc_opt_in": False},
        {"fbc_opt_in": False},
        {"fbc_opt_in": False},
    ]

    mock_get_deprecation_list.side_effect = [["bundle1", "bundle2"], ["bundle3"]]

    mock_run_entrypoint.side_effect = [
        [{"ocp_version": "4.5"}, {"ocp_version": "4.6"}],
    ]
    iib_results = [
        IIBRes(
            "some-registry.com/index-image:5",
            "some-registry.com/index-image@sha256:a1a1",
            ["v4.5-test-hotfix-RHBA-1234-4567-3"],
        ),
        IIBRes(
            "some-registry.com/index-image:6",
            "some-registry.com/index-image@sha256:b2b2",
            ["v4.6-test-hotfix-RHBA-1234-4567-3"],
        ),
    ]
    mock_add_bundles.side_effect = iib_results
    pusher = operator_pusher.OperatorPusher(
        [operator_push_item_hotfix_invalid_origin], "3", target_settings
    )

    with pytest.raises(ValueError) as exc:
        results = pusher.build_index_images()
    assert str(exc.value) == "Cannot push hotfixes without an advisory"


@mock.patch("pubtools._quay.operator_pusher.ContainerImagePusher.run_tag_images")
@mock.patch("pubtools._quay.operator_pusher.OperatorPusher.iib_add_bundles")
@mock.patch("pubtools._quay.operator_pusher.run_entrypoint")
@mock.patch("pubtools._quay.operator_pusher.OperatorPusher.get_deprecation_list")
@mock.patch("pubtools._quay.operator_pusher.pyxis_get_repo_metadata")
def test_push_operators_prerelease_invalid_origin(
    mock_get_repo_metadata,
    mock_get_deprecation_list,
    mock_run_entrypoint,
    mock_add_bundles,
    mock_run_tag_images,
    target_settings,
    operator_push_item_prerelease_invalid_origin,
    fake_cert_key_paths,
):
    mock_get_repo_metadata.side_effect = [
        {"fbc_opt_in": False},
        {"fbc_opt_in": False},
        {"fbc_opt_in": False},
    ]

    mock_get_deprecation_list.side_effect = [["bundle1", "bundle2"], ["bundle3"]]

    mock_run_entrypoint.side_effect = [
        [{"ocp_version": "4.5"}, {"ocp_version": "4.6"}],
    ]
    iib_results = [
        IIBRes(
            "some-registry.com/index-image:5",
            "some-registry.com/index-image@sha256:a1a1",
            ["v4.5-test-hotfix-RHBA-1234-4567-3"],
        ),
        IIBRes(
            "some-registry.com/index-image:6",
            "some-registry.com/index-image@sha256:b2b2",
            ["v4.6-test-hotfix-RHBA-1234-4567-3"],
        ),
    ]
    mock_add_bundles.side_effect = iib_results
    pusher = operator_pusher.OperatorPusher(
        [operator_push_item_prerelease_invalid_origin], "3", target_settings
    )

    with pytest.raises(ValueError) as exc:
        results = pusher.build_index_images()
    assert str(exc.value) == "Cannot push pre release without an advisory"


@mock.patch("pubtools._quay.push_docker.QuayClient")
@mock.patch("pubtools._quay.push_docker.QuayApiClient")
@mock.patch("pubtools._quay.operator_pusher.run_entrypoint")
@mock.patch("pubtools._quay.operator_pusher.pyxis_get_repo_metadata")
def test_get_existing_index_images(
    mock_get_repo_metadata,
    mock_run_entrypoint,
    mock_quay_client,
    mock_quay_api,
    target_settings,
    operator_push_item_ok,
    manifest_list_data,
    fake_cert_key_paths,
):
    mock_get_repo_metadata.side_effect = [
        {"fbc_opt_in": False},
        {"fbc_opt_in": False},
    ]
    mock_run_entrypoint.return_value = [{"ocp_version": "4.5"}, {"ocp_version": "4.6"}]
    mock_quay_client.get_manifest.return_value = manifest_list_data

    pusher = operator_pusher.OperatorPusher([operator_push_item_ok], "3", target_settings)
    existing_index_images = pusher.get_existing_index_images(mock_quay_client)
    mock_quay_client.get_manifest.assert_has_calls(
        [
            mock.call(
                "quay.io/some-namespace/operators----index-image:v4.5",
                media_type="application/vnd.docker.distribution.manifest.list.v2+json",
            )
        ]
    )
    mock_quay_client.get_manifest.assert_has_calls(
        [
            mock.call(
                "quay.io/some-namespace/operators----index-image:v4.6",
                media_type="application/vnd.docker.distribution.manifest.list.v2+json",
            )
        ]
    )
    assert sorted(existing_index_images) == [
        (
            "sha256:146ab6fa7ba3ab4d154b09c1c5522e4966ecd071bf23d1ba3df6c8b9fc33f8cb",
            "v4.5",
            "operators/index-image",
        ),
        (
            "sha256:146ab6fa7ba3ab4d154b09c1c5522e4966ecd071bf23d1ba3df6c8b9fc33f8cb",
            "v4.6",
            "operators/index-image",
        ),
        (
            "sha256:2e8f38a0a8d2a450598430fa70c7f0b53aeec991e76c3e29c63add599b4ef7ee",
            "v4.5",
            "operators/index-image",
        ),
        (
            "sha256:2e8f38a0a8d2a450598430fa70c7f0b53aeec991e76c3e29c63add599b4ef7ee",
            "v4.6",
            "operators/index-image",
        ),
        (
            "sha256:496fb0ff2057c79254c9dc6ba999608a98219c5c93142569a547277c679e532c",
            "v4.5",
            "operators/index-image",
        ),
        (
            "sha256:496fb0ff2057c79254c9dc6ba999608a98219c5c93142569a547277c679e532c",
            "v4.6",
            "operators/index-image",
        ),
        (
            "sha256:b3f9218fb5839763e62e52ee6567fe331aa1f3c644f9b6f232ff23959257acf9",
            "v4.5",
            "operators/index-image",
        ),
        (
            "sha256:b3f9218fb5839763e62e52ee6567fe331aa1f3c644f9b6f232ff23959257acf9",
            "v4.6",
            "operators/index-image",
        ),
        (
            "sha256:bbef1f46572d1f33a92b53b0ba0ed5a1d09dab7ffe64be1ae3ae66e76275eabd",
            "v4.5",
            "operators/index-image",
        ),
        (
            "sha256:bbef1f46572d1f33a92b53b0ba0ed5a1d09dab7ffe64be1ae3ae66e76275eabd",
            "v4.6",
            "operators/index-image",
        ),
    ]


@mock.patch("pubtools._quay.push_docker.QuayClient")
@mock.patch("pubtools._quay.push_docker.QuayApiClient")
@mock.patch("pubtools._quay.operator_pusher.run_entrypoint")
@mock.patch("pubtools._quay.operator_pusher.pyxis_get_repo_metadata")
def test_get_existing_index_images_fbc(
    mock_get_repo_metadata,
    mock_run_entrypoint,
    mock_quay_client,
    mock_quay_api,
    target_settings,
    operator_push_item_ok,
    manifest_list_data,
    fake_cert_key_paths,
):
    mock_get_repo_metadata.side_effect = [
        {"fbc_opt_in": True},
        {"fbc_opt_in": False},
    ]
    mock_run_entrypoint.return_value = [{"ocp_version": "4.11"}, {"ocp_version": "4.12"}]
    mock_quay_client.get_manifest.return_value = manifest_list_data

    pusher = operator_pusher.OperatorPusher([operator_push_item_ok], "3", target_settings)
    existing_index_images = pusher.get_existing_index_images(mock_quay_client)
    print(mock_quay_client.mock_calls)
    assert sorted(existing_index_images) == [
        (
            "sha256:146ab6fa7ba3ab4d154b09c1c5522e4966ecd071bf23d1ba3df6c8b9fc33f8cb",
            "v4.12",
            "operators/index-image",
        ),
        (
            "sha256:2e8f38a0a8d2a450598430fa70c7f0b53aeec991e76c3e29c63add599b4ef7ee",
            "v4.12",
            "operators/index-image",
        ),
        (
            "sha256:496fb0ff2057c79254c9dc6ba999608a98219c5c93142569a547277c679e532c",
            "v4.12",
            "operators/index-image",
        ),
        (
            "sha256:b3f9218fb5839763e62e52ee6567fe331aa1f3c644f9b6f232ff23959257acf9",
            "v4.12",
            "operators/index-image",
        ),
        (
            "sha256:bbef1f46572d1f33a92b53b0ba0ed5a1d09dab7ffe64be1ae3ae66e76275eabd",
            "v4.12",
            "operators/index-image",
        ),
    ]


@mock.patch("pubtools._quay.push_docker.QuayClient")
@mock.patch("pubtools._quay.push_docker.QuayApiClient")
@mock.patch("pubtools._quay.operator_pusher.run_entrypoint")
@mock.patch("pubtools._quay.operator_pusher.pyxis_get_repo_metadata")
def test_get_existing_index_images_raises_401(
    mock_get_repo_metadata,
    mock_run_entrypoint,
    mock_quay_client,
    mock_quay_api,
    target_settings,
    operator_push_item_ok,
    manifest_list_data,
    fake_cert_key_paths,
):
    mock_get_repo_metadata.side_effect = [
        {"fbc_opt_in": False},
        {"fbc_opt_in": False},
        {"fbc_opt_in": False},
    ]
    mock_run_entrypoint.return_value = [{"ocp_version": "4.5"}, {"ocp_version": "4.6"}]
    mock_quay_client.get_manifest.side_effect = [
        requests.exceptions.HTTPError(response=mock.Mock(status_code=401)),
        manifest_list_data,
    ]

    pusher = operator_pusher.OperatorPusher([operator_push_item_ok], "3", target_settings)
    existing_index_images = pusher.get_existing_index_images(mock_quay_client)
    mock_quay_client.get_manifest.assert_has_calls(
        [
            mock.call(
                "quay.io/some-namespace/operators----index-image:v4.5",
                media_type="application/vnd.docker.distribution.manifest.list.v2+json",
            )
        ]
    )
    mock_quay_client.get_manifest.assert_has_calls(
        [
            mock.call(
                "quay.io/some-namespace/operators----index-image:v4.6",
                media_type="application/vnd.docker.distribution.manifest.list.v2+json",
            )
        ]
    )

    assert sorted(existing_index_images) == [
        (
            "sha256:146ab6fa7ba3ab4d154b09c1c5522e4966ecd071bf23d1ba3df6c8b9fc33f8cb",
            "v4.6",
            "operators/index-image",
        ),
        (
            "sha256:2e8f38a0a8d2a450598430fa70c7f0b53aeec991e76c3e29c63add599b4ef7ee",
            "v4.6",
            "operators/index-image",
        ),
        (
            "sha256:496fb0ff2057c79254c9dc6ba999608a98219c5c93142569a547277c679e532c",
            "v4.6",
            "operators/index-image",
        ),
        (
            "sha256:b3f9218fb5839763e62e52ee6567fe331aa1f3c644f9b6f232ff23959257acf9",
            "v4.6",
            "operators/index-image",
        ),
        (
            "sha256:bbef1f46572d1f33a92b53b0ba0ed5a1d09dab7ffe64be1ae3ae66e76275eabd",
            "v4.6",
            "operators/index-image",
        ),
    ]


@mock.patch("pubtools._quay.push_docker.QuayClient")
@mock.patch("pubtools._quay.push_docker.QuayApiClient")
@mock.patch("pubtools._quay.operator_pusher.run_entrypoint")
@mock.patch("pubtools._quay.operator_pusher.pyxis_get_repo_metadata")
def test_get_existing_index_images_raises_500(
    mock_get_repo_metadata,
    mock_run_entrypoint,
    mock_quay_client,
    mock_quay_api,
    target_settings,
    operator_push_item_ok,
    manifest_list_data,
    fake_cert_key_paths,
):
    mock_get_repo_metadata.side_effect = [
        {"fbc_opt_in": False},
        {"fbc_opt_in": False},
        {"fbc_opt_in": False},
    ]
    mock_run_entrypoint.return_value = [{"ocp_version": "4.5"}, {"ocp_version": "4.6"}]
    mock_quay_client.get_manifest.side_effect = [
        requests.exceptions.HTTPError("500", response=mock.Mock(status_code=500)),
        manifest_list_data,
    ]

    pusher = operator_pusher.OperatorPusher([operator_push_item_ok], "3", target_settings)
    with pytest.raises(requests.exceptions.HTTPError, match=".*500.*"):
        existing_index_images = pusher.get_existing_index_images(mock_quay_client)


@mock.patch("pubtools._quay.operator_pusher.OperatorPusher.iib_add_bundles")
@mock.patch("pubtools._quay.operator_pusher.run_entrypoint")
@mock.patch("pubtools._quay.operator_pusher.OperatorPusher.get_deprecation_list")
@mock.patch("pubtools._quay.operator_pusher.pyxis_get_repo_metadata")
def test_push_operators_fbc_opted_in(
    mock_get_repo_metadata,
    mock_get_deprecation_list,
    mock_run_entrypoint,
    mock_add_bundles,
    target_settings,
    operator_push_item_fbc,
    operator_push_item_different_version,
    fake_cert_key_paths,
):
    mock_get_repo_metadata.side_effect = [
        {"fbc_opt_in": True},
        {"fbc_opt_in": True},
        {"fbc_opt_in": True},
    ]

    mock_get_deprecation_list.side_effect = [["bundle1", "bundle2"], ["bundle3"], []]

    mock_run_entrypoint.side_effect = [
        [{"ocp_version": "4.6"}, {"ocp_version": "4.7"}],
        [{"ocp_version": "4.11"}],
    ]
    iib_results = [
        IIBRes(
            "some-registry.com/ns/index-image:6",
            "some-registry.com/ns/iib@sha256:b2b2",
            ["v4.6-3"],
        ),
        IIBRes(
            "some-registry.com/ns/index-image:7",
            "some-registry.com/ns/iib@sha256:c3c3",
            ["v4.7-3"],
        ),
    ]
    mock_add_bundles.side_effect = iib_results
    pusher = operator_pusher.OperatorPusher(
        [operator_push_item_fbc, operator_push_item_different_version], "3", target_settings
    )

    results = pusher.build_index_images()

    assert mock_get_deprecation_list.call_count == 3
    assert mock_get_deprecation_list.call_args_list[0] == mock.call("v4.11")
    assert mock_get_deprecation_list.call_args_list[1] == mock.call("v4.6")
    assert mock_get_deprecation_list.call_args_list[2] == mock.call("v4.7")

    assert results == {
        "v4.6-v4.6": {
            "destination_tags": ["v4.6"],
            "iib_result": iib_results[0],
            "signing_keys": ["some-key"],
        },
        "v4.7-v4.7": {
            "destination_tags": ["v4.7"],
            "iib_result": iib_results[1],
            "signing_keys": ["some-key"],
        },
    }
    assert mock_add_bundles.call_count == 2


@mock.patch("pubtools._quay.operator_pusher.OperatorPusher.iib_add_bundles")
@mock.patch("pubtools._quay.operator_pusher.run_entrypoint")
@mock.patch("pubtools._quay.operator_pusher.OperatorPusher.get_deprecation_list")
@mock.patch("pubtools._quay.operator_pusher.pyxis_get_repo_metadata")
def test_push_operators_fbc_opted_in_inconsistent(
    mock_get_repo_metadata,
    mock_get_deprecation_list,
    mock_run_entrypoint,
    mock_add_bundles,
    target_settings,
    operator_push_item_fbc,
    operator_push_item_different_version,
    fake_cert_key_paths,
):
    """Expected push failure as item repositories are inconsistently opted in.
    Failure happens on push item and therefore no iib operations should happen."""

    mock_get_repo_metadata.side_effect = [
        {"fbc_opt_in": True},
        {"fbc_opt_in": False},
        {"fbc_opt_in": True},
    ]

    mock_get_deprecation_list.side_effect = [["bundle1", "bundle2"], ["bundle3"], []]

    mock_run_entrypoint.side_effect = [
        [{"ocp_version": "4.6"}, {"ocp_version": "4.7"}],
        [{"ocp_version": "4.10"}],
    ]
    iib_results = [
        IIBRes(
            "some-registry.com/ns/index-image:8",
            "some-registry.com/ns/iib@sha256:c3c3",
            ["v4.10-4"],
        )
    ]
    mock_add_bundles.side_effect = iib_results
    pusher = operator_pusher.OperatorPusher(
        [operator_push_item_fbc, operator_push_item_different_version], "3", target_settings
    )

    results = pusher.build_index_images()

    assert mock_get_deprecation_list.call_count == 0

    assert results == {}
    assert mock_add_bundles.call_count == 0
    assert operator_push_item_fbc.state == "NOTPUSHED"


@mock.patch("pubtools._quay.operator_pusher.OperatorPusher.iib_add_bundles")
@mock.patch("pubtools._quay.operator_pusher.run_entrypoint")
@mock.patch("pubtools._quay.operator_pusher.OperatorPusher.get_deprecation_list")
@mock.patch("pubtools._quay.operator_pusher.pyxis_get_repo_metadata")
def test_push_operators_fbc_4_9_to_4_11_opted_in(
    mock_get_repo_metadata,
    mock_get_deprecation_list,
    mock_run_entrypoint,
    mock_add_bundles,
    target_settings,
    operator_push_item_fbc,
    operator_push_item_different_version,
    fake_cert_key_paths,
):
    """Expected push failure as item is opted in fbc and targeted to < 4.13 and  >= 4.13.
    Failure happens on push item and therefore no iib operations should happen."""

    mock_get_repo_metadata.side_effect = [
        {"fbc_opt_in": True},
        {"fbc_opt_in": True},
        {"fbc_opt_in": True},
    ]

    mock_get_deprecation_list.side_effect = [["bundle1", "bundle2"], ["bundle3"], [], [], []]

    mock_run_entrypoint.side_effect = [
        [
            {"ocp_version": "4.9"},
            {"ocp_version": "4.10"},
            {"ocp_version": "4.11"},
        ],
        [{"ocp_version": "4.6"}, {"ocp_version": "4.7"}],
    ]

    iib_results = [{}, {}]

    mock_add_bundles.side_effect = iib_results
    pusher = operator_pusher.OperatorPusher(
        [operator_push_item_fbc, operator_push_item_different_version], "3", target_settings
    )

    results = pusher.build_index_images()

    assert results == {}

    assert mock_add_bundles.call_count == 0
    assert operator_push_item_fbc.state == "INVALIDFILE"



import logging
import mock
import pytest
import os
import json
import subprocess

from pubtools._quay import security_manifest_pusher
from pubtools._quay.exceptions import ManifestTypeError

TEST_DATA_PATH = os.path.join(os.path.dirname(__file__), "test_data")


@mock.patch("pubtools._quay.security_manifest_pusher.QuayApiClient")
@mock.patch("pubtools._quay.security_manifest_pusher.QuayClient")
def test_init(
    mock_quay_client, mock_quay_api_client, target_settings, container_multiarch_push_item
):
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )

    assert pusher.push_items == [container_multiarch_push_item]
    assert pusher.target_settings == target_settings
    assert pusher.quay_host == "quay.io"
    mock_quay_client.assert_not_called()
    mock_quay_api_client.assert_not_called()

    assert pusher.src_quay_client == mock_quay_client.return_value
    mock_quay_client.assert_called_once_with("src-quay-user", "src-quay-pass", "quay.io")
    assert pusher.dest_quay_api_client == mock_quay_api_client.return_value
    mock_quay_api_client.assert_called_once_with("dest-quay-token", "quay.io")


@mock.patch("subprocess.run")
def test_cosign_get_security_manifest(mock_run, target_settings, container_multiarch_push_item):
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )
    mock_run.return_value.returncode = 0

    res = pusher.cosign_get_security_manifest("quay.io/org/repo@sha256:abcdef", "/temp/file.txt")
    assert res
    mock_run.assert_called_once_with(
        [
            "cosign",
            "download",
            "sbom",
            "quay.io/org/repo@sha256:abcdef",
            "--output-file",
            "/temp/file.txt",
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        text=True,
    )


@mock.patch("subprocess.run")
def test_cosign_get_security_manifest_normal_error(
    mock_run, target_settings, container_multiarch_push_item
):
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )
    mock_run.return_value.returncode = 1
    mock_run.return_value.stdout = """
    Error: no sbom attached to reference
    main.go:74: error during command execution: no sbom attached to reference
    """

    res = pusher.cosign_get_security_manifest("quay.io/org/repo@sha256:abcdef", "/temp/file.txt")
    assert not res


@mock.patch("time.sleep")
@mock.patch("subprocess.run")
def test_cosign_get_security_manifest_unusual_error(
    mock_run, mock_sleep, target_settings, container_multiarch_push_item
):
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )
    mock_run.return_value.returncode = 1
    mock_run.return_value.stdout = """
    Error: unusual error
    main.go:74: error during command execution: unusual error
    """

    with pytest.raises(RuntimeError, match=".*has failed with an error.*"):
        pusher.cosign_get_security_manifest("quay.io/org/repo@sha256:abcdef", "/temp/file.txt")

    assert mock_run.call_count == 4


@mock.patch("subprocess.run")
def test_cosign_get_existing_attestation(mock_run, target_settings, container_multiarch_push_item):
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )
    mock_run.return_value.returncode = 0

    res = pusher.cosign_get_existing_attestation("quay.io/org/repo@sha256:abcdef", "/temp/file.txt")
    assert res
    mock_run.assert_called_once_with(
        [
            "cosign",
            "verify-attestation",
            "--key",
            "path/to/key.pub",
            "quay.io/org/repo@sha256:abcdef",
            "--output-file",
            "/temp/file.txt",
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        text=True,
    )


@mock.patch("subprocess.run")
def test_cosign_get_existing_attestation_rekor_url(
    mock_run, target_settings, container_multiarch_push_item
):
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )
    mock_run.return_value.returncode = 0

    res = pusher.cosign_get_existing_attestation(
        "quay.io/org/repo@sha256:abcdef", "/temp/file.txt", "https://some-rekor.com"
    )
    assert res
    mock_run.assert_called_once_with(
        [
            "cosign",
            "verify-attestation",
            "--rekor-url=https://some-rekor.com",
            "--key",
            "path/to/key.pub",
            "quay.io/org/repo@sha256:abcdef",
            "--output-file",
            "/temp/file.txt",
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        text=True,
    )


@mock.patch("subprocess.run")
def test_cosign_get_existing_attestation_disable_rekor(
    mock_run, target_settings, container_multiarch_push_item
):
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )
    mock_run.return_value.returncode = 0

    res = pusher.cosign_get_existing_attestation(
        "quay.io/org/repo@sha256:abcdef", "/temp/file.txt", "https://some-rekor.com", True
    )
    assert res
    mock_run.assert_called_once_with(
        [
            "cosign",
            "verify-attestation",
            "--insecure-ignore-tlog=true",
            "--key",
            "path/to/key.pub",
            "quay.io/org/repo@sha256:abcdef",
            "--output-file",
            "/temp/file.txt",
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        text=True,
    )


@mock.patch("subprocess.run")
def test_cosign_get_existing_attestation_normal_error(
    mock_run, target_settings, container_multiarch_push_item
):
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )
    mock_run.return_value.returncode = 1
    mock_run.return_value.stdout = (
        "Error: no matching attestations: \n"
        "main.go:74: error during command execution: no matching attestations: \n"
    )

    res = pusher.cosign_get_existing_attestation("quay.io/org/repo@sha256:abcdef", "/temp/file.txt")
    assert not res


@mock.patch("time.sleep")
@mock.patch("subprocess.run")
def test_cosign_get_existing_attestation_unusual_error(
    mock_run, mock_sleep, target_settings, container_multiarch_push_item
):
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )
    mock_run.return_value.returncode = 1
    mock_run.return_value.stdout = """
    Error: no matching attestations: unusual error
    main.go:74: error during command execution: no matching attestations: unusual error
    """

    with pytest.raises(RuntimeError, match=".*has failed with an error.*"):
        pusher.cosign_get_existing_attestation("quay.io/org/repo@sha256:abcdef", "/temp/file.txt")

    assert mock_run.call_count == 4


@mock.patch("subprocess.run")
def test_cosign_attest_security_manifest(mock_run, target_settings, container_multiarch_push_item):
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )
    mock_run.return_value.returncode = 0

    pusher.cosign_attest_security_manifest("/temp/file.txt", "quay.io/org/repo@sha256:abcdef")
    mock_run.assert_called_once_with(
        [
            "cosign",
            "attest",
            "--predicate",
            "/temp/file.txt",
            "--key",
            "path/to/key.key",
            "-y",
            "quay.io/org/repo@sha256:abcdef",
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        text=True,
    )


@mock.patch("subprocess.run")
def test_cosign_attest_security_manifest_rekor_url(
    mock_run, target_settings, container_multiarch_push_item
):
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )
    mock_run.return_value.returncode = 0

    pusher.cosign_attest_security_manifest(
        "/temp/file.txt", "quay.io/org/repo@sha256:abcdef", "https://some-rekor.com"
    )
    mock_run.assert_called_once_with(
        [
            "cosign",
            "attest",
            "--rekor-url=https://some-rekor.com",
            "--predicate",
            "/temp/file.txt",
            "--key",
            "path/to/key.key",
            "-y",
            "quay.io/org/repo@sha256:abcdef",
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        text=True,
    )


@mock.patch("subprocess.run")
def test_cosign_attest_security_manifest_disable_rekor(
    mock_run, target_settings, container_multiarch_push_item
):
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )
    mock_run.return_value.returncode = 0

    pusher.cosign_attest_security_manifest(
        "/temp/file.txt", "quay.io/org/repo@sha256:abcdef", "https://some-rekor.com", True
    )
    mock_run.assert_called_once_with(
        [
            "cosign",
            "attest",
            "--tlog-upload=false",
            "--predicate",
            "/temp/file.txt",
            "--key",
            "path/to/key.key",
            "-y",
            "quay.io/org/repo@sha256:abcdef",
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        text=True,
    )


@mock.patch("time.sleep")
@mock.patch("subprocess.run")
def test_cosign_attest_security_manifest_err(
    mock_run, mock_sleep, target_settings, container_multiarch_push_item
):
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )
    mock_run.return_value.returncode = 1

    with pytest.raises(RuntimeError, match="Creating attestation to image.*"):
        pusher.cosign_attest_security_manifest("/temp/file.txt", "quay.io/org/repo@sha256:abcdef")

    assert mock_run.call_count == 4


@mock.patch("uuid.uuid4")
@mock.patch("subprocess.run")
@mock.patch(
    "pubtools._quay.security_manifest_pusher.open",
    mock.mock_open(read_data="quay.io/org/repo:sha256-abcdef.att\n"),
)
def test_cosign_triangulate_image(
    mock_run, mock_uuid, target_settings, container_multiarch_push_item
):
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )
    mock_run.return_value.returncode = 0
    mock_uuid.return_value.hex = "abcd"

    ret = pusher.cosign_triangulate_image("quay.io/org/repo@sha256:abcdef", "/temp/")
    mock_run.assert_called_once_with(
        [
            "cosign",
            "triangulate",
            "--type=attestation",
            "quay.io/org/repo@sha256:abcdef",
            "--output-file",
            "/temp/attestation_reference_abcd.json",
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        text=True,
    )
    assert ret == "quay.io/org/repo:sha256-abcdef.att"


@mock.patch("uuid.uuid4")
@mock.patch("subprocess.run")
@mock.patch(
    "pubtools._quay.security_manifest_pusher.open",
    mock.mock_open(read_data="quay.io/org/repo:sha256-abcdef.sig\n"),
)
def test_cosign_triangulate_image_nondefault_type(
    mock_run, mock_uuid, target_settings, container_multiarch_push_item
):
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )
    mock_run.return_value.returncode = 0
    mock_uuid.return_value.hex = "abcd"

    ret = pusher.cosign_triangulate_image("quay.io/org/repo@sha256:abcdef", "/temp/", "signature")
    mock_run.assert_called_once_with(
        [
            "cosign",
            "triangulate",
            "--type=signature",
            "quay.io/org/repo@sha256:abcdef",
            "--output-file",
            "/temp/signature_reference_abcd.json",
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        text=True,
    )
    assert ret == "quay.io/org/repo:sha256-abcdef.sig"


@mock.patch("uuid.uuid4")
@mock.patch("subprocess.run")
@mock.patch(
    "pubtools._quay.security_manifest_pusher.open",
    mock.mock_open(read_data="quay.io/org/repo:sha256-abcdef.att\n"),
)
def test_cosign_triangulate_image_error(
    mock_run, mock_uuid, target_settings, container_multiarch_push_item
):
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )
    mock_run.return_value.returncode = 1
    mock_uuid.return_value.hex = "abcd"

    with pytest.raises(RuntimeError, match="Triangulating attestation image to image.*"):
        pusher.cosign_triangulate_image("quay.io/org/repo@sha256:abcdef", "/temp/")


@mock.patch("uuid.uuid4")
@mock.patch("subprocess.run")
@mock.patch(
    "pubtools._quay.security_manifest_pusher.open",
    mock.mock_open(read_data="quay.io/org/repo:sha256-abcdef.att\n"),
)
def test_cosign_triangulate_image_unknown_type(
    mock_run, mock_uuid, target_settings, container_multiarch_push_item
):
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )
    mock_run.return_value.returncode = 0
    mock_uuid.return_value.hex = "abcd"

    with pytest.raises(ValueError, match="Image type 'wrong' needs to be one of.*"):
        pusher.cosign_triangulate_image("quay.io/org/repo@sha256:abcdef", "/temp/", "wrong")


def test_get_security_manifest_from_attestation(target_settings, container_multiarch_push_item):
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )
    with open(os.path.join(TEST_DATA_PATH, "security_manifest.json"), "r") as f:
        expected_security_manifest = json.load(f)

    security_manifest = pusher.get_security_manifest_from_attestation(
        os.path.join(TEST_DATA_PATH, "raw_attestation.json")
    )

    assert security_manifest == expected_security_manifest


def test_security_manifest_get_products(target_settings, container_multiarch_push_item):
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )
    with open(os.path.join(TEST_DATA_PATH, "security_manifest_products.json"), "r") as f:
        security_manifest = json.load(f)

    products = pusher.security_manifest_get_products(security_manifest)
    assert products == {"product1", "product2"}


def test_get_destination_repos(target_settings, container_multiarch_push_item):
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )

    dest_repos = pusher.get_destination_repos(container_multiarch_push_item)
    assert dest_repos == ["quay.io/some-namespace/target----repo"]


@mock.patch("uuid.uuid4")
@mock.patch("json.dump")
@mock.patch("json.load")
@mock.patch("pubtools._quay.security_manifest_pusher.open", mock.mock_open())
def test_security_manifest_remove_incompleteness_reasons(
    mock_load, mock_dump, mock_uuid, target_settings, container_multiarch_push_item
):
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )
    mock_load.return_value = {"specVersion": "1.4", "incompleteness_reasons": []}
    mock_uuid.return_value.hex = "abcd"

    res = pusher.security_manifest_remove_incompleteness_reasons(__file__)
    assert res == os.path.join(os.path.dirname(__file__), "sanitized_security_manifest_abcd.json")
    mock_load.assert_called_once()
    mock_dump.assert_called_once()
    assert mock_dump.call_args_list[0][0][0] == {"specVersion": "1.4"}


@mock.patch("uuid.uuid4")
@mock.patch("json.dump")
@mock.patch("json.load")
@mock.patch("pubtools._quay.security_manifest_pusher.open", mock.mock_open())
def test_security_manifest_add_products(
    mock_load, mock_dump, mock_uuid, target_settings, container_multiarch_push_item
):
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )
    mock_load.return_value = {}
    mock_uuid.return_value.hex = "abcd"

    res = pusher.security_manifest_add_products(__file__, {"product1", "product2"})
    assert res == os.path.join(os.path.dirname(__file__), "full_security_manifest_abcd.json")
    mock_load.assert_called_once()
    mock_dump.assert_called_once()
    mock_dump.call_args_list[0][0][0]["properties"].sort(key=lambda d: d["value"])
    assert mock_dump.call_args_list[0][0][0] == {
        "properties": [
            {"name": "product", "value": "product1"},
            {"name": "product", "value": "product2"},
        ]
    }


@mock.patch("uuid.uuid4")
@mock.patch("json.dump")
@mock.patch("pubtools._quay.security_manifest_pusher.open", mock.mock_open())
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher."
    "get_security_manifest_from_attestation"
)
def test_extract_security_manifest_from_attestation(
    mock_get_sec_manifest, mock_dump, mock_uuid, target_settings, container_multiarch_push_item
):
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )
    mock_get_sec_manifest.return_value = {"specVersion": "1.4", "incompleteness_reasons": []}
    mock_uuid.return_value.hex = "abcd"

    path = pusher.extract_security_manifest_from_attestation("attestation.json")
    assert path == "extracted_security_manifest_abcd.json"

    mock_get_sec_manifest.assert_called_once_with("attestation.json")
    mock_dump.assert_called_once()


@mock.patch("pubtools._quay.security_manifest_pusher.QuayApiClient")
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher." "cosign_triangulate_image"
)
def test_delete_existing_attestation(
    mock_triangulate, mock_api_client, target_settings, container_multiarch_push_item
):
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )
    mock_triangulate.return_value = "quay.io/org/repo:sha256-abcdef.att"

    pusher.delete_existing_attestation("quay.io/org/repo@sha256:abcdef", "/temp")
    mock_triangulate.assert_called_once_with("quay.io/org/repo@sha256:abcdef", "/temp")
    mock_api_client.assert_called_once_with("dest-quay-token", "quay.io")


@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher.delete_existing_attestation"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher.cosign_attest_security_manifest"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher.security_manifest_add_products"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher"
    ".security_manifest_remove_incompleteness_reasons"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher."
    "get_security_manifest_from_attestation"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher.cosign_get_existing_attestation"
)
@mock.patch("uuid.uuid4")
def test_merge_and_push_security_manifest(
    mock_uuid,
    mock_get_attestation,
    mock_get_manifest,
    mock_remove_incompleteness,
    mock_add_products,
    mock_attest,
    mock_delete_attestation,
    target_settings,
    container_multiarch_push_item,
):
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )
    with open(os.path.join(TEST_DATA_PATH, "security_manifest_products.json"), "r") as f:
        security_manifest = json.load(f)

    digest_manifest = security_manifest_pusher.DigestSecurityManifest("abcdef", "path/to/manifest")
    mock_get_attestation.return_value = True

    mock_get_manifest.return_value = security_manifest
    mock_remove_incompleteness.return_value = "/path/to/sanitized/manifest.json"

    mock_uuid.return_value.hex = "abcd"
    mock_add_products.return_value = "/path/to/final/manifest.json"

    pusher.merge_and_push_security_manifest(
        container_multiarch_push_item, digest_manifest, ["quay.io/org/repo"], "/temp/temp_path"
    )

    mock_get_attestation.assert_called_once_with(
        "quay.io/org/repo@abcdef",
        "/temp/temp_path/attestation_abcd.json",
        "https://some-rekor.com",
        False,
    )
    mock_get_manifest.assert_called_once_with("/temp/temp_path/attestation_abcd.json")
    mock_remove_incompleteness.assert_called_once_with("path/to/manifest")
    mock_add_products.assert_called_once_with(
        "/path/to/sanitized/manifest.json", {"new-product", "product1", "product2"}
    )
    mock_delete_attestation.assert_called_once_with("quay.io/org/repo@abcdef", "/temp/temp_path")
    mock_attest.assert_called_once_with(
        "/path/to/final/manifest.json", "quay.io/org/repo@abcdef", "https://some-rekor.com", False
    )


@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher.delete_existing_attestation"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher.cosign_attest_security_manifest"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher.security_manifest_add_products"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher"
    ".security_manifest_remove_incompleteness_reasons"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher."
    "get_security_manifest_from_attestation"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher.cosign_get_existing_attestation"
)
@mock.patch("uuid.uuid4")
def test_merge_and_push_security_manifest_no_existing_attestation_also_disable_rekor(
    mock_uuid,
    mock_get_attestation,
    mock_get_manifest,
    mock_remove_incompleteness,
    mock_add_products,
    mock_attest,
    mock_delete_attestation,
    target_settings,
    container_multiarch_push_item,
):
    target_settings["cosign_sbom_skip_verify_rekor"] = True
    target_settings["cosign_sbom_skip_upload_rekor"] = True
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )
    with open(os.path.join(TEST_DATA_PATH, "security_manifest_products.json"), "r") as f:
        security_manifest = json.load(f)

    digest_manifest = security_manifest_pusher.DigestSecurityManifest("abcdef", "path/to/manifest")
    mock_get_attestation.return_value = False

    mock_get_manifest.return_value = security_manifest
    mock_remove_incompleteness.return_value = "/path/to/sanitized/manifest.json"

    mock_uuid.return_value.hex = "abcd"
    mock_add_products.return_value = "/path/to/final/manifest.json"

    pusher.merge_and_push_security_manifest(
        container_multiarch_push_item, digest_manifest, ["quay.io/org/repo"], "/temp/temp_path"
    )

    mock_get_attestation.assert_called_once_with(
        "quay.io/org/repo@abcdef",
        "/temp/temp_path/attestation_abcd.json",
        "https://some-rekor.com",
        True,
    )
    mock_get_manifest.assert_not_called()
    mock_delete_attestation.assert_not_called()
    mock_remove_incompleteness.assert_called_once_with("path/to/manifest")
    mock_add_products.assert_called_once_with("/path/to/sanitized/manifest.json", {"new-product"})
    mock_attest.assert_called_once_with(
        "/path/to/final/manifest.json", "quay.io/org/repo@abcdef", "https://some-rekor.com", True
    )


@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher.delete_existing_attestation"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher.cosign_attest_security_manifest"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher.security_manifest_add_products"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher"
    ".security_manifest_remove_incompleteness_reasons"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher."
    "get_security_manifest_from_attestation"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher.cosign_get_existing_attestation"
)
@mock.patch("uuid.uuid4")
def test_merge_and_push_security_manifest_already_pushed(
    mock_uuid,
    mock_get_attestation,
    mock_get_manifest,
    mock_remove_incompleteness,
    mock_add_products,
    mock_attest,
    mock_delete_attestation,
    target_settings,
    container_multiarch_push_item,
):
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )
    with open(os.path.join(TEST_DATA_PATH, "security_manifest_products.json"), "r") as f:
        security_manifest = json.load(f)
    security_manifest["properties"] = [{"name": "product", "value": "new-product"}]

    digest_manifest = security_manifest_pusher.DigestSecurityManifest("abcdef", "path/to/manifest")
    mock_get_attestation.return_value = True

    mock_get_manifest.return_value = security_manifest
    mock_remove_incompleteness.return_value = "/path/to/sanitized/manifest.json"

    mock_uuid.return_value.hex = "abcd"
    mock_add_products.return_value = "/path/to/final/manifest.json"

    pusher.merge_and_push_security_manifest(
        container_multiarch_push_item, digest_manifest, ["quay.io/org/repo"], "/temp/temp_path"
    )

    mock_get_attestation.assert_called_once_with(
        "quay.io/org/repo@abcdef",
        "/temp/temp_path/attestation_abcd.json",
        "https://some-rekor.com",
        False,
    )
    mock_get_manifest.assert_called_once_with("/temp/temp_path/attestation_abcd.json")
    mock_delete_attestation.assert_not_called()
    mock_remove_incompleteness.assert_not_called()
    mock_add_products.assert_not_called()
    mock_attest.assert_not_called()


@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher.delete_existing_attestation"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher.cosign_attest_security_manifest"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher.security_manifest_add_products"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher"
    ".security_manifest_remove_incompleteness_reasons"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher."
    "get_security_manifest_from_attestation"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher.cosign_get_existing_attestation"
)
@mock.patch("uuid.uuid4")
def test_merge_and_push_security_manifest_no_product_existing_attestation(
    mock_uuid,
    mock_get_attestation,
    mock_get_manifest,
    mock_remove_incompleteness,
    mock_add_products,
    mock_attest,
    mock_delete_attestation,
    target_settings,
    container_multiarch_push_item,
):
    container_multiarch_push_item.metadata.pop("product_name")
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )
    with open(os.path.join(TEST_DATA_PATH, "security_manifest_products.json"), "r") as f:
        security_manifest = json.load(f)

    digest_manifest = security_manifest_pusher.DigestSecurityManifest("abcdef", "path/to/manifest")
    mock_get_attestation.return_value = True
    mock_remove_incompleteness.return_value = "/path/to/sanitized/manifest.json"

    mock_get_manifest.return_value = security_manifest

    mock_uuid.return_value.hex = "abcd"
    mock_add_products.return_value = "/path/to/final/manifest.json"

    pusher.merge_and_push_security_manifest(
        container_multiarch_push_item, digest_manifest, ["quay.io/org/repo"], "/temp/temp_path"
    )

    mock_get_attestation.assert_called_once_with(
        "quay.io/org/repo@abcdef",
        "/temp/temp_path/attestation_abcd.json",
        "https://some-rekor.com",
        False,
    )
    mock_get_manifest.assert_called_once_with("/temp/temp_path/attestation_abcd.json")
    mock_add_products.assert_not_called()
    mock_delete_attestation.assert_not_called()
    mock_remove_incompleteness.assert_not_called()
    mock_attest.assert_not_called()


@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher.delete_existing_attestation"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher.cosign_attest_security_manifest"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher.security_manifest_add_products"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher"
    ".security_manifest_remove_incompleteness_reasons"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher."
    "get_security_manifest_from_attestation"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher.cosign_get_existing_attestation"
)
@mock.patch("uuid.uuid4")
def test_merge_and_push_security_manifest_no_product_no_existing_attestation(
    mock_uuid,
    mock_get_attestation,
    mock_get_manifest,
    mock_remove_incompleteness,
    mock_add_products,
    mock_attest,
    mock_delete_attestation,
    target_settings,
    container_multiarch_push_item,
):
    container_multiarch_push_item.metadata.pop("product_name")
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )
    with open(os.path.join(TEST_DATA_PATH, "security_manifest_products.json"), "r") as f:
        security_manifest = json.load(f)

    digest_manifest = security_manifest_pusher.DigestSecurityManifest("abcdef", "path/to/manifest")
    mock_get_attestation.return_value = False

    mock_get_manifest.return_value = security_manifest
    mock_remove_incompleteness.return_value = "/path/to/sanitized/manifest.json"

    mock_uuid.return_value.hex = "abcd"
    mock_add_products.return_value = "/path/to/final/manifest.json"

    pusher.merge_and_push_security_manifest(
        container_multiarch_push_item, digest_manifest, ["quay.io/org/repo"], "/temp/temp_path"
    )

    mock_get_attestation.assert_called_once_with(
        "quay.io/org/repo@abcdef",
        "/temp/temp_path/attestation_abcd.json",
        "https://some-rekor.com",
        False,
    )
    mock_get_manifest.assert_not_called()
    mock_add_products.assert_not_called()
    mock_remove_incompleteness.assert_called_once_with("path/to/manifest")
    mock_delete_attestation.assert_not_called()
    mock_attest.assert_called_once_with(
        "/path/to/sanitized/manifest.json",
        "quay.io/org/repo@abcdef",
        "https://some-rekor.com",
        False,
    )


@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher."
    "extract_security_manifest_from_attestation"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher.delete_existing_attestation"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher.cosign_attest_security_manifest"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher.cosign_get_existing_attestation"
)
@mock.patch("uuid.uuid4")
@mock.patch("pubtools._quay.security_manifest_pusher.QuayClient")
def test_push_manifest_list_security_manifests(
    mock_quay_client,
    mock_uuid,
    mock_get_attestation,
    mock_attest,
    mock_delete_attestation,
    mock_extract_sec_manifest,
    target_settings,
    manifest_list_data,
    container_multiarch_push_item,
):
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )
    mock_get_attestation.return_value = True
    mock_quay_client.return_value.get_manifest_digest.return_value = "sha256:abcdef"
    mock_quay_client.return_value.get_manifest.return_value = manifest_list_data
    mock_uuid.return_value.hex = "abcd"
    mock_extract_sec_manifest.return_value = "extracted_security_manifest_defg.json"

    pusher.push_manifest_list_security_manifests(
        container_multiarch_push_item,
        "/some",
    )

    mock_quay_client.return_value.get_manifest_digest.assert_called_once_with(
        "quay.io/some-namespace/target----repo:latest-test-tag",
        mock_quay_client.return_value.MANIFEST_LIST_TYPE,
    )
    mock_quay_client.return_value.get_manifest.assert_called_once_with(
        "quay.io/some-namespace/target----repo:latest-test-tag",
        media_type=mock_quay_client.return_value.MANIFEST_LIST_TYPE,
    )

    assert mock_get_attestation.call_count == 6
    assert mock_get_attestation.call_args_list[0] == mock.call(
        "quay.io/some-namespace/target----repo@"
        "sha256:2e8f38a0a8d2a450598430fa70c7f0b53aeec991e76c3e29c63add599b4ef7ee",
        "/some/attestation_abcd.json",
        "https://some-rekor.com",
        False,
    )
    assert mock_get_attestation.call_args_list[1] == mock.call(
        "quay.io/some-namespace/target----repo@"
        "sha256:b3f9218fb5839763e62e52ee6567fe331aa1f3c644f9b6f232ff23959257acf9",
        "/some/attestation_abcd.json",
        "https://some-rekor.com",
        False,
    )
    assert mock_get_attestation.call_args_list[2] == mock.call(
        "quay.io/some-namespace/target----repo@"
        "sha256:496fb0ff2057c79254c9dc6ba999608a98219c5c93142569a547277c679e532c",
        "/some/attestation_abcd.json",
        "https://some-rekor.com",
        False,
    )
    assert mock_get_attestation.call_args_list[3] == mock.call(
        "quay.io/some-namespace/target----repo@"
        "sha256:146ab6fa7ba3ab4d154b09c1c5522e4966ecd071bf23d1ba3df6c8b9fc33f8cb",
        "/some/attestation_abcd.json",
        "https://some-rekor.com",
        False,
    )
    assert mock_get_attestation.call_args_list[4] == mock.call(
        "quay.io/some-namespace/target----repo@"
        "sha256:bbef1f46572d1f33a92b53b0ba0ed5a1d09dab7ffe64be1ae3ae66e76275eabd",
        "/some/attestation_abcd.json",
        "https://some-rekor.com",
        False,
    )
    assert mock_get_attestation.call_args_list[5] == mock.call(
        "quay.io/some-namespace/target----repo@sha256:abcdef",
        "/some/attestation_abcd.json",
        "https://some-rekor.com",
        False,
    )
    assert mock_extract_sec_manifest.call_count == 5
    assert mock_extract_sec_manifest.call_args_list[0] == mock.call("/some/attestation_abcd.json")
    mock_delete_attestation.assert_called_once_with(
        "quay.io/some-namespace/target----repo@sha256:abcdef", "/some"
    )

    assert mock_attest.call_count == 5
    assert mock_attest.call_args_list[0] == mock.call(
        "extracted_security_manifest_defg.json",
        "quay.io/some-namespace/target----repo@sha256:abcdef",
        "https://some-rekor.com",
        False,
    )


@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher."
    "extract_security_manifest_from_attestation"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher.delete_existing_attestation"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher.cosign_attest_security_manifest"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher.cosign_get_existing_attestation"
)
@mock.patch("uuid.uuid4")
@mock.patch("pubtools._quay.security_manifest_pusher.QuayClient")
def test_push_manifest_list_security_manifests_arch_att_not_exist(
    mock_quay_client,
    mock_uuid,
    mock_get_attestation,
    mock_attest,
    mock_delete_attestation,
    mock_extract_sec_manifest,
    target_settings,
    manifest_list_data,
    container_multiarch_push_item,
):
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )
    mock_get_attestation.side_effect = [True, True, False, True, True, True]
    mock_quay_client.return_value.get_manifest_digest.return_value = "sha256:abcdef"
    mock_quay_client.return_value.get_manifest.return_value = manifest_list_data
    mock_uuid.return_value.hex = "abcd"
    mock_extract_sec_manifest.return_value = "extracted_security_manifest_defg.json"

    pusher.push_manifest_list_security_manifests(
        container_multiarch_push_item,
        "/some",
    )

    mock_quay_client.return_value.get_manifest_digest.assert_called_once_with(
        "quay.io/some-namespace/target----repo:latest-test-tag",
        mock_quay_client.return_value.MANIFEST_LIST_TYPE,
    )
    mock_quay_client.return_value.get_manifest.assert_called_once_with(
        "quay.io/some-namespace/target----repo:latest-test-tag",
        media_type=mock_quay_client.return_value.MANIFEST_LIST_TYPE,
    )

    assert mock_get_attestation.call_count == 6
    assert mock_extract_sec_manifest.call_count == 4
    assert mock_extract_sec_manifest.call_args_list[0] == mock.call("/some/attestation_abcd.json")
    mock_delete_attestation.assert_called_once_with(
        "quay.io/some-namespace/target----repo@sha256:abcdef", "/some"
    )
    assert mock_attest.call_count == 4


@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher.cosign_get_security_manifest"
)
@mock.patch("uuid.uuid4")
@mock.patch("pubtools._quay.security_manifest_pusher.QuayClient")
def test_get_source_item_security_manifests(
    mock_quay_client,
    mock_uuid,
    mock_get_security_manifest,
    target_settings,
    container_source_push_item,
):
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_source_push_item], target_settings
    )

    mock_quay_client.return_value.get_manifest_digest.return_value = "sha256:abcdef"
    mock_uuid.return_value.hex = "abcd"
    mock_get_security_manifest.return_value = ["abc"]

    res = pusher.get_source_item_security_manifests(container_source_push_item, "/temp")

    assert res == [
        security_manifest_pusher.DigestSecurityManifest(
            "sha256:abcdef", "/temp/security_manifest_source_abcd.json"
        )
    ]
    mock_quay_client.return_value.get_manifest_digest.assert_called_once_with(
        "some-registry/src/repo:2"
    )
    mock_get_security_manifest.assert_called_once_with(
        "some-registry/src/repo:2", "/temp/security_manifest_source_abcd.json"
    )


@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher.cosign_get_security_manifest"
)
@mock.patch("uuid.uuid4")
@mock.patch("pubtools._quay.security_manifest_pusher.QuayClient")
def test_get_source_item_security_manifests_no_manifest(
    mock_quay_client,
    mock_uuid,
    mock_get_security_manifest,
    target_settings,
    container_source_push_item,
):
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_source_push_item], target_settings
    )

    mock_quay_client.return_value.get_manifest_digest.return_value = "sha256:abcdef"
    mock_uuid.return_value.hex = "abcd"
    mock_get_security_manifest.return_value = []

    res = pusher.get_source_item_security_manifests(container_source_push_item, "/temp")
    assert res == []


@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher.cosign_get_security_manifest"
)
@mock.patch("uuid.uuid4")
@mock.patch("pubtools._quay.security_manifest_pusher.QuayClient")
def test_get_multiarch_item_security_manifests(
    mock_quay_client,
    mock_uuid,
    mock_get_manifest,
    target_settings,
    container_multiarch_push_item,
    src_manifest_list,
    caplog,
):
    caplog.set_level(logging.ERROR)
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )

    mock_quay_client.return_value.get_manifest.return_value = src_manifest_list
    mock_uuid.return_value.hex = "abcd"
    mock_get_manifest.return_value = True

    res = pusher.get_multiarch_item_security_manifests(container_multiarch_push_item, "/temp")
    assert res == [
        security_manifest_pusher.DigestSecurityManifest(
            "sha256:1111111111", "/temp/security_manifest_arm64_abcd.json"
        ),
        security_manifest_pusher.DigestSecurityManifest(
            "sha256:2222222222", "/temp/security_manifest_armhfp_abcd.json"
        ),
        security_manifest_pusher.DigestSecurityManifest(
            "sha256:3333333333", "/temp/security_manifest_ppc64le_abcd.json"
        ),
        security_manifest_pusher.DigestSecurityManifest(
            "sha256:5555555555", "/temp/security_manifest_amd64_abcd.json"
        ),
    ]
    mock_quay_client.return_value.get_manifest.assert_called_once()
    assert mock_get_manifest.call_count == 4
    assert mock_get_manifest.call_args_list[0] == mock.call(
        "some-registry/src/repo@sha256:1111111111", "/temp/security_manifest_arm64_abcd.json"
    )
    assert mock_get_manifest.call_args_list[1] == mock.call(
        "some-registry/src/repo@sha256:2222222222", "/temp/security_manifest_armhfp_abcd.json"
    )
    assert mock_get_manifest.call_args_list[2] == mock.call(
        "some-registry/src/repo@sha256:3333333333", "/temp/security_manifest_ppc64le_abcd.json"
    )
    assert mock_get_manifest.call_args_list[3] == mock.call(
        "some-registry/src/repo@sha256:5555555555", "/temp/security_manifest_amd64_abcd.json"
    )

    assert caplog.messages == []


@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher.cosign_get_security_manifest"
)
@mock.patch("uuid.uuid4")
@mock.patch("pubtools._quay.security_manifest_pusher.QuayClient")
def test_get_multiarch_item_security_manifests_not_present(
    mock_quay_client,
    mock_uuid,
    mock_get_manifest,
    target_settings,
    container_multiarch_push_item,
    src_manifest_list,
    caplog,
):
    caplog.set_level(logging.ERROR)
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )

    mock_quay_client.return_value.get_manifest.return_value = src_manifest_list
    mock_uuid.return_value.hex = "abcd"
    mock_get_manifest.return_value = False

    res = pusher.get_multiarch_item_security_manifests(container_multiarch_push_item, "/temp")
    assert res == []
    mock_quay_client.return_value.get_manifest.assert_called_once()
    assert mock_get_manifest.call_count == 4
    assert caplog.messages == []


@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher.cosign_get_security_manifest"
)
@mock.patch("uuid.uuid4")
@mock.patch("pubtools._quay.security_manifest_pusher.QuayClient")
def test_get_multiarch_item_security_manifests_some_present(
    mock_quay_client,
    mock_uuid,
    mock_get_manifest,
    target_settings,
    container_multiarch_push_item,
    src_manifest_list,
    caplog,
):
    caplog.set_level(logging.ERROR)
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )

    mock_quay_client.return_value.get_manifest.return_value = src_manifest_list
    mock_uuid.return_value.hex = "abcd"
    mock_get_manifest.side_effect = [True, False, True, False]

    res = pusher.get_multiarch_item_security_manifests(container_multiarch_push_item, "/temp")
    assert res == [
        security_manifest_pusher.DigestSecurityManifest(
            "sha256:1111111111", "/temp/security_manifest_arm64_abcd.json"
        ),
        security_manifest_pusher.DigestSecurityManifest(
            "sha256:3333333333", "/temp/security_manifest_ppc64le_abcd.json"
        ),
    ]
    mock_quay_client.return_value.get_manifest.assert_called_once()
    assert mock_get_manifest.call_count == 4
    assert caplog.messages == [
        "Only some architectures of multiarch image some-registry/src/repo:1 have a "
        "security manifest"
    ]


@mock.patch("tempfile.TemporaryDirectory")
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher."
    "push_manifest_list_security_manifests"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher."
    "merge_and_push_security_manifest"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher."
    "get_source_item_security_manifests"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher."
    "get_multiarch_item_security_manifests"
)
@mock.patch("pubtools._quay.security_manifest_pusher.QuayClient")
def test_push_item_security_manifest_source_image(
    mock_quay_client,
    mock_get_multiarch,
    mock_get_source,
    mock_merge_and_push,
    mock_push_ml_security_manifests,
    mock_tmpdir,
    target_settings,
    container_source_push_item,
):
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_source_push_item], target_settings
    )
    retval = security_manifest_pusher.DigestSecurityManifest(
        "sha256:abcdef", "/temp/security_manifest_source_abcd.json"
    )
    mock_get_source.return_value = [retval]
    mock_quay_client.return_value.get_manifest.side_effect = ManifestTypeError("not found")
    mock_tmpdir.return_value.__enter__.return_value = "tmp/path"

    pusher.push_item_security_manifests(container_source_push_item)
    mock_quay_client.return_value.get_manifest.assert_called_once()
    mock_get_source.assert_called_once_with(container_source_push_item, "tmp/path")
    mock_get_multiarch.assert_not_called()
    mock_merge_and_push.assert_called_once_with(
        container_source_push_item,
        retval,
        ["quay.io/some-namespace/target----repo"],
        mock_tmpdir.return_value.__enter__.return_value,
    )
    mock_push_ml_security_manifests.assert_not_called()


@mock.patch("tempfile.TemporaryDirectory")
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher."
    "push_manifest_list_security_manifests"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher."
    "merge_and_push_security_manifest"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher."
    "get_source_item_security_manifests"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher."
    "get_multiarch_item_security_manifests"
)
@mock.patch("pubtools._quay.security_manifest_pusher.QuayClient")
def test_push_item_security_manifest_multiarch_image(
    mock_quay_client,
    mock_get_multiarch,
    mock_get_source,
    mock_merge_and_push,
    mock_push_ml_security_manifests,
    mock_tmpdir,
    target_settings,
    container_multiarch_push_item,
    manifest_list_data,
):
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )
    retval = [
        security_manifest_pusher.DigestSecurityManifest(
            "sha256:1111111111", "/temp/security_manifest_arm64_abcd.json"
        ),
        security_manifest_pusher.DigestSecurityManifest(
            "sha256:3333333333", "/temp/security_manifest_ppc64le_abcd.json"
        ),
    ]
    mock_get_multiarch.return_value = retval
    mock_quay_client.return_value.get_manifest.return_value = manifest_list_data
    mock_tmpdir.return_value.__enter__.return_value = "tmp/path"

    pusher.push_item_security_manifests(container_multiarch_push_item)
    mock_quay_client.return_value.get_manifest.assert_called_once()
    mock_get_source.assert_not_called()
    mock_get_multiarch.assert_called_once_with(container_multiarch_push_item, "tmp/path")
    assert mock_merge_and_push.call_count == 2
    assert mock_merge_and_push.call_args_list[0] == mock.call(
        container_multiarch_push_item,
        retval[0],
        ["quay.io/some-namespace/target----repo"],
        mock_tmpdir.return_value.__enter__.return_value,
    )
    assert mock_merge_and_push.call_args_list[1] == mock.call(
        container_multiarch_push_item,
        retval[1],
        ["quay.io/some-namespace/target----repo"],
        mock_tmpdir.return_value.__enter__.return_value,
    )

    mock_push_ml_security_manifests.assert_called_once_with(
        container_multiarch_push_item,
        "tmp/path",
    )


@mock.patch("tempfile.TemporaryDirectory")
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher."
    "push_manifest_list_security_manifests"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher."
    "merge_and_push_security_manifest"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher."
    "get_source_item_security_manifests"
)
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher."
    "get_multiarch_item_security_manifests"
)
@mock.patch("pubtools._quay.security_manifest_pusher.QuayClient")
def test_push_item_security_manifest_v2s1_image(
    mock_quay_client,
    mock_get_multiarch,
    mock_get_source,
    mock_merge_and_push,
    mock_push_ml_security_manifests,
    mock_tmpdir,
    target_settings,
    container_multiarch_push_item,
):
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item], target_settings
    )
    retval = security_manifest_pusher.DigestSecurityManifest(
        "sha256:abcdef", "/temp/security_manifest_source_abcd.json"
    )
    mock_get_source.return_value = [retval]
    mock_quay_client.return_value.get_manifest.side_effect = ManifestTypeError("not found")
    mock_tmpdir.return_value.__enter__.return_value.name = "tmp/path"

    pusher.push_item_security_manifests(container_multiarch_push_item)
    mock_quay_client.return_value.get_manifest.assert_called_once()
    mock_get_source.assert_not_called()
    mock_get_multiarch.assert_not_called()
    mock_merge_and_push.assert_not_called()
    mock_push_ml_security_manifests.assert_not_called()


@mock.patch("pubtools._quay.security_manifest_pusher.LocalExecutor")
@mock.patch(
    "pubtools._quay.security_manifest_pusher.SecurityManifestPusher.push_item_security_manifests"
)
def test_push_security_manifests(
    mock_push_item_security_manifests,
    mock_local_executor,
    target_settings,
    container_multiarch_push_item,
    container_source_push_item,
):
    pusher = security_manifest_pusher.SecurityManifestPusher(
        [container_multiarch_push_item, container_source_push_item], target_settings
    )
    pusher.push_security_manifests()

    mock_local_executor.return_value.__enter__.return_value.skopeo_login.assert_called_once_with(
        "quay.io", "dest-quay-user", "dest-quay-pass"
    )
    assert mock_push_item_security_manifests.call_count == 2



import logging
import argparse
from typing import Any, Optional, List, Dict

from pubtools.pluggy import pm, task_context

from .image_untagger import ImageUntagger
from .utils.misc import setup_arg_parser, add_args_env_variables

LOG = logging.getLogger("pubtools.quay")

UNTAG_IMAGES_ARGS = {
    ("--reference",): {
        "help": "Image reference to untag. Must be specified by tag. Multiple can be specified.",
        "required": True,
        "type": str,
        "action": "append",
    },
    ("--remove-last",): {
        "help": "Whether to remove a tag even if it's the last reference of some image.",
        "required": False,
        "type": bool,
    },
    ("--quay-api-token",): {
        "help": "OAuth token for Quay REST API.",
        "required": False,
        "type": str,
        "env_variable": "QUAY_API_TOKEN",
    },
    ("--quay-user",): {
        "help": "Username for Quay login.",
        "required": False,
        "type": str,
    },
    ("--quay-password",): {
        "help": "Password for Quay. Can be specified by env variable QUAY_PASSWORD.",
        "required": False,
        "type": str,
        "env_variable": "QUAY_PASSWORD",
    },
}


def construct_kwargs(args: argparse.Namespace) -> Dict[Any, Any]:
    """
    Construct a kwargs dictionary based on the entered command line arguments.

    Args:
        args (argparse.Namespace):
            Parsed command line arguments.

    Returns (dict):
        Keyword arguments for the 'untag_images' function.
    """
    kwargs = args.__dict__

    # in args.__dict__ unspecified bool values have 'None' instead of 'False'
    for name, attributes in UNTAG_IMAGES_ARGS.items():
        if attributes["type"] is bool:
            bool_var = name[0].lstrip("-").replace("-", "_")
            if kwargs[bool_var] is None:
                kwargs[bool_var] = False

    # some exceptions have to be remapped
    kwargs["references"] = kwargs.pop("reference")

    return kwargs


def verify_untag_images_args(
    references: List[str], quay_user: Optional[str], quay_password: Optional[str]
) -> None:
    """
    Verify the presence and correctness of input parameters.

    Args:
        references ([str]):
            List of image references to untag.
        quay_user (str):
            Quay username for Docker HTTP API.
        quay_password (str):
            Quay password for Docker HTTP API.
    """
    for reference in references:
        if "@" in reference:
            raise ValueError("All references must be specified via tag, not digest")

    if (quay_user and not quay_password) or (quay_password and not quay_user):
        raise ValueError("Both user and password must be present when attempting to log in.")


def untag_images(
    references: List[str],
    quay_api_token: str,
    remove_last: bool = False,
    quay_user: Optional[str] = None,
    quay_password: Optional[str] = None,
) -> None:
    """
    Untag images from Quay.

    Args:
        references ([str]):
            List of image references to untag.
        quay_api_token (str):
            OAuth token for authentication of Quay REST API.
        remove_last (bool):
            Whether to remove a tag when it's the last reference of an image (in that repo).
        quay_user (str):
            Quay username for Docker HTTP API.
        quay_password (str):
            Quay password for Docker HTTP API.
    """
    verify_untag_images_args(references, quay_user, quay_password)

    LOG.info("Started untagging operation with the following references: {0}".format(references))
    untagger = ImageUntagger(references, quay_api_token, remove_last, quay_user, quay_password)
    lost_images = untagger.untag_images()

    LOG.info("Untagging operation succeeded")
    pm.hook.quay_images_untagged(untag_refs=sorted(references), lost_refs=sorted(lost_images))


def setup_args() -> argparse.ArgumentParser:
    """Set up argparser without extra parameters, this method is used for auto doc generation."""
    return setup_arg_parser(UNTAG_IMAGES_ARGS)


def untag_images_main(sysargs: Optional[List[str]] = None) -> None:
    """Entrypoint for untagging images."""
    logging.basicConfig(level=logging.INFO)

    parser = setup_args()
    if sysargs:
        args = parser.parse_args(sysargs[1:])
    else:
        args = parser.parse_args()  # pragma: no cover"
    args = add_args_env_variables(args, UNTAG_IMAGES_ARGS)

    if not args.quay_api_token:
        raise ValueError("--quay-api-token must be specified")

    kwargs = construct_kwargs(args)

    with task_context():
        untag_images(**kwargs)
